import atexit
import logging
import os.path
import tempfile
from typing import Any, Dict, List, Optional, Tuple, Union

import dask
import dask.dataframe as dd
import pandas as pd
from dask.cache import Cache
from dask.utils import parse_bytes
from distributed import Client, LocalCluster

from biome.text.environment import (
    DEFAULT_DASK_CACHE_SIZE,
    ENV_DASK_CACHE_SIZE,
)

__LOGGER = logging.getLogger(__name__)

__DASK_CLIENT = None


def configure_dask_cluster(
    address: str = "local", n_workers: int = 1, worker_memory: Union[str, int] = "1GB"
) -> Optional[Client]:
    """Creates a dask client (with a LocalCluster if needed)

    Parameters
    ----------
    address
        The cluster address. If "local" try to connect to a local cluster listening the 8786 port.
        If no cluster listening, creates a new LocalCluster
    n_workers
        The number of cluster workers (only a new "local" cluster generation)
    worker_memory
        The memory reserved for local workers

    Returns
    -------
    A new dask Client

    """
    global __DASK_CLIENT  # pylint: disable=global-statement

    if __DASK_CLIENT:
        return __DASK_CLIENT

    def create_dask_client(
        dask_cluster: str, cache_size: Optional[int], workers: int, worker_mem: int
    ) -> Client:
        if cache_size:
            cache = Cache(cache_size)
            cache.register()

        if dask_cluster != "local":
            return Client(dask_cluster)

        try:
            return Client("localhost:8786", timeout=5)
        except OSError:
            cluster = _create_local_cluster(workers, worker_mem)
            return Client(cluster)

    def _create_local_cluster(workers: int, worker_mem: int):
        """Creates a dask local cluster instance"""
        dask.config.set(
            {
                "distributed.worker.memory": dict(
                    target=0.95, spill=False, pause=False, terminate=False
                )
            }
        )
        dask_data = os.path.join(os.getcwd(), ".dask")
        os.makedirs(dask_data, exist_ok=True)
        worker_space = tempfile.mkdtemp(dir=dask_data)
        cluster = LocalCluster(
            n_workers=workers,
            threads_per_worker=1,
            asynchronous=False,
            scheduler_port=8786,  # TODO configurable
            processes=False,
            memory_limit=worker_mem,
            silence_logs=logging.ERROR,
            local_directory=worker_space,
        )
        return cluster

    dask_cluster = address
    dask_cache_size = os.environ.get(ENV_DASK_CACHE_SIZE, DEFAULT_DASK_CACHE_SIZE)

    if isinstance(worker_memory, str):
        worker_memory = parse_bytes(worker_memory)

    __DASK_CLIENT = create_dask_client(
        dask_cluster, dask_cache_size, workers=n_workers, worker_mem=worker_memory
    )

    return __DASK_CLIENT


@atexit.register
def close_dask_client():
    global __DASK_CLIENT  # pylint: disable=global-statement

    try:
        __DASK_CLIENT.close(timeout=10)
        __DASK_CLIENT.cluster.close(timeout=10)
    except Exception as err:  # pylint: disable=broad-except
        __LOGGER.debug(err)
    finally:
        __DASK_CLIENT = None


def extension_from_path(path: Union[str, List[str]]) -> str:
    """Helper method to get file extension

    Parameters
    ----------
    path
        A string or a list of strings.
        If it is a list, the first entry is taken.

    Returns
    -------
    extension
        File extension
    """
    if isinstance(path, str):
        path = [path]

    _, extension = os.path.splitext(path[0])

    return extension.lower()[1:]  # skip first char, which is a dot


def make_paths_relative(yaml_dirname: str, cfg_dict: Dict, path_keys: List[str] = None):
    """Helper method to convert file system paths relative to the yaml config file,
    to paths relative to the current path.

    It will recursively cycle through `cfg_dict` if it is nested.

    Parameters
    ----------
    yaml_dirname
        Dirname to the yaml config file (as obtained by `os.path.dirname`.
    cfg_dict
        The config dictionary extracted from the yaml file.
    path_keys
        If not None, it will only try to modify the `cfg_dict` values corresponding to the `path_keys`.
    """
    for key, value in cfg_dict.items():
        if isinstance(value, dict):
            make_paths_relative(yaml_dirname, value, path_keys)

        if path_keys and key not in path_keys:
            continue

        if is_relative_file_system_path(value):  # returns False if value is not a str
            cfg_dict[key] = os.path.join(yaml_dirname, value)

        # cover lists as well
        if isinstance(value, list):
            cfg_dict[key] = [
                os.path.join(yaml_dirname, path)
                if is_relative_file_system_path(path)
                else path
                for path in value
            ]


def is_relative_file_system_path(string: str) -> bool:
    """Helper method to check if a string is a relative file system path.

    Parameters
    ----------
    string
        The string to be checked.

    Returns
    -------
    bool
        Whether the string is a relative file system path or not.
        If string is not type(str), return False.
    """
    if not isinstance(string, str):
        return False
    # we require the files to have a file name extension ... ¯\_(ツ)_/¯
    if not os.path.isdir(string) and not extension_from_path(string):
        return False
    # check if a domain name
    if string.lower().startswith(
        ("http://", "https://", "ftp://", "sftp://", "s3://", "hdfs://", "gs://")
    ):
        return False
    # check if an absolute path
    if os.path.isabs(string):
        return False
    return True


def _dict_to_list(row: List[Dict]) -> Optional[dict]:
    """ Converts a list of structured data into a dict of list, where every dict key
        is the list aggregation for every key in original dict

        For example:

        l = [{"name": "Frank", "lastName":"Ocean"},{"name":"Oliver","lastName":"Sacks"]
        _dict_to_list(l)
        {"name":["Frank","Oliver"], "lastName":["Ocean", "Sacks"]}
    """
    try:
        for row_i in row:
            if isinstance(row_i, list):
                row = row_i
        return pd.DataFrame(row).to_dict(orient="list")
    except (ValueError, TypeError):
        return None


def _columns_analysis(
    data_frame: pd.DataFrame,
) -> Tuple[List[str], List[str], List[str]]:
    dicts = []
    lists = []
    unmodified = []

    def is_list_of_structured_data(elem) -> bool:
        if isinstance(elem, list):
            for elem_i in elem:
                if isinstance(elem_i, (dict, list)):
                    return True
        return False

    for column in data_frame.columns:
        column_data = data_frame[column].dropna()
        element = column_data.iloc[0] if not column_data.empty else None

        current_list = unmodified
        if isinstance(element, dict):
            current_list = dicts
        elif is_list_of_structured_data(element):
            current_list = lists
        current_list.append(column)

    return dicts, lists, unmodified


def flatten_dask_dataframe(data_frame: dd.DataFrame) -> dd.DataFrame:
    """
    Flatten an dataframe adding nested values as new columns
    and dropping the old ones
    Parameters
    ----------
    data_frame
        The original dask DataFrame

    Returns
    -------

    A new Dataframe with flatten content

    """
    # We must materialize some data for compound the new flatten DataFrame
    meta_flatten = flatten_dataframe(data_frame.head(1))

    def _flatten_stage(data_frame_i: pd.DataFrame) -> pd.DataFrame:
        new_df = flatten_dataframe(data_frame_i)
        for column in new_df.columns:
            # we append the new columns to the original dataframe
            data_frame_i[column] = new_df[column]

        return data_frame_i

    data_frame = data_frame.map_partitions(
        _flatten_stage,
        meta={**data_frame.dtypes.to_dict(), **meta_flatten.dtypes.to_dict()},
    )
    return data_frame[meta_flatten.columns]


def flatten_dataframe(data_frame: pd.DataFrame) -> pd.DataFrame:
    dict_columns, list_columns, unmodified_columns = _columns_analysis(data_frame)

    if len(data_frame.columns) == len(unmodified_columns):
        return data_frame

    dfs = []
    for column in list_columns:
        column_df = pd.DataFrame(
            data=[data for data in data_frame[column].apply(_dict_to_list) if data],
            index=data_frame.index,
        )
        column_df.columns = [
            f"{column}.*.{column_df_column}" for column_df_column in column_df.columns
        ]
        dfs.append(column_df)

    for column in dict_columns:
        column_df = pd.DataFrame(
            data=[data if data else {} for data in data_frame[column]],
            index=data_frame.index,
        )
        column_df.columns = [
            f"{column}.{column_df_column}" for column_df_column in column_df.columns
        ]
        dfs.append(column_df)

    flatten = flatten_dataframe(pd.concat(dfs, axis=1))
    return pd.concat([data_frame[unmodified_columns], flatten], axis=1)



import os

from biome.text.data import DataSource
from tests import DaskSupportTest, RESOURCES_PATH

FILES_PATH = os.path.join(RESOURCES_PATH, "data")


class ExcelDataSourceTest(DaskSupportTest):
    def test_read_excel(self):
        file_path = os.path.join(FILES_PATH, "test.xlsx")

        datasource = DataSource(format="xlsx", source=file_path)
        data_frame = datasource.to_dataframe().compute()

        assert len(data_frame) > 0
        self.assertTrue("path" in data_frame.columns)



import inspect
import os
import os.path
import re
from inspect import Parameter
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Type

import torch
import yaml
from allennlp.data import TextFieldTensors
from elasticsearch import Elasticsearch

from . import environment
from .features import CharFeatures, WordFeatures

_INVALID_TAG_CHARACTERS = re.compile(r"[^-/\w\.]")


def yaml_to_dict(filepath: str) -> Dict[str, Any]:
    """Loads a yaml file into a data dictionary"""
    with open(filepath) as yaml_content:
        config = yaml.safe_load(yaml_content)
    return config


def get_compatible_doc_type(client: Elasticsearch) -> str:
    """
    Find a compatible name for doc type by checking the cluster info
    Parameters
    ----------
    client
        The elasticsearch client

    Returns
    -------
        A compatible name for doc type in function of cluster version
    """

    es_version = int(client.info()["version"]["number"].split(".")[0])
    return "_doc" if es_version >= 6 else "doc"


def get_env_cuda_device() -> int:
    """Gets the cuda device from an environment variable.

    This is necessary to activate a GPU if available

    Returns
    -------
    cuda_device
        The integer number of the CUDA device
    """
    cuda_device = int(os.getenv(environment.CUDA_DEVICE, "-1"))
    return cuda_device


def update_method_signature(signature: inspect.Signature, to_method):
    """Updates signature to method"""

    def wrapper(*args, **kwargs):
        return to_method(*args, **kwargs)

    wrapper.__signature__ = signature
    return wrapper


def isgeneric(class_type: Type) -> bool:
    """Checks if a class type is a generic type (List[str] or Union[str, int]"""
    return hasattr(class_type, "__origin__")


def is_running_on_notebook() -> bool:
    """Checks if code is running inside a jupyter notebook"""
    try:
        import IPython

        return IPython.get_ipython().has_trait("kernel")
    except (AttributeError, NameError, ModuleNotFoundError):
        return False


def split_signature_params_by_predicate(
    signature_function: Callable, predicate: Callable
) -> Tuple[List[Parameter], List[Parameter]]:
    """Splits parameters signature by defined boolean predicate function"""
    signature = inspect.signature(signature_function)
    parameters = list(
        filter(
            lambda p: p.name != "self"
            and p.kind not in [Parameter.VAR_KEYWORD, Parameter.VAR_POSITIONAL],
            signature.parameters.values(),
        )
    )
    matches_group = list(filter(lambda p: predicate(p), parameters))
    non_matches_group = list(filter(lambda p: not predicate(p), parameters))

    return matches_group, non_matches_group


def clean_metric_name(name):
    if not name:
        return name
    new_name = _INVALID_TAG_CHARACTERS.sub("_", name)
    new_name = new_name.lstrip("/")
    return new_name


def get_word_tokens_ids_from_text_field_tensors(
    text_field_tensors: TextFieldTensors,
) -> Optional[torch.Tensor]:
    """
    Given a text field tensor structure, tries to extract word features related tensors

    Parameters
    ----------
    text_field_tensors: The incoming record text field tensors dictionary

    Returns
    -------

    `WordFeatures` related tensors if enable
    """
    word_features_tensors = text_field_tensors.get(WordFeatures.namespace)
    if not word_features_tensors:
        return None

    for argument_name, tensor in word_features_tensors.items():
        if argument_name in ["tokens", "token_ids", "input_ids"]:
            return tensor


def get_char_tokens_ids_from_text_field_tensors(
    text_field_tensors: TextFieldTensors,
) -> Optional[torch.Tensor]:
    """
    Given a text field tensor structure, tries to extract character features related tensors

    See `TokenCharactersIndexer.tokens_to_indices` for more info

    Parameters
    ----------
    text_field_tensors: The incoming record text field tensors dictionary

    Returns
    -------

    `CharFeatures` related tensors if enable
    """
    char_features_tensors = text_field_tensors.get(CharFeatures.namespace)
    if not char_features_tensors:
        return None

    for argument_name, tensor in char_features_tensors.items():
        if argument_name in ["token_characters"]:
            return tensor


def save_dict_as_yaml(dictionary: dict, path: str) -> str:
    """Save a cfg dict to path as yaml

    Parameters
    ----------
    dictionary
        Dictionary to be saved
    path
        Filesystem location where the yaml file will be saved

    Returns
    -------
    path
        Location of the yaml file
    """
    dir_name = os.path.dirname(path)
    # Prevent current workdir relative routes
    # `save_dict_as_yaml("just_here.yml")
    if dir_name:
        os.makedirs(dir_name, exist_ok=True)

    with open(path, "w") as yml_file:
        yaml.dump(dictionary, yml_file, default_flow_style=False, allow_unicode=True)

    return path


def get_full_class_name(the_class: Type) -> str:
    """Given a type class return the full qualified class name """
    # o.__module__ + "." + o.__class__.__qualname__ is an example in
    # this context of H.L. Mencken's "neat, plausible, and wrong."
    # Python makes no guarantees as to whether the __module__ special
    # attribute is defined, so we take a more circumspect approach.
    # Alas, the module name is explicitly excluded from __qualname__
    # in Python 3.

    module = the_class.__module__
    if module is None or module == str.__class__.__module__:
        return the_class.__name__  # Avoid reporting __builtin__
    else:
        return module + "." + the_class.__name__


def stringify(value: Any) -> Any:
    """Creates an equivalent data structure representing data values as string"""
    if isinstance(value, str):
        return value
    if isinstance(value, dict):
        return {key: stringify(value) for key, value in value.items()}
    if isinstance(value, Iterable):
        return [stringify(v) for v in value]
    return str(value)



from biome.text.text_cleaning import DefaultTextCleaning
from biome.text.tokenizer import Tokenizer

html_text = """
        <!DOCTYPE html>
        <html>
        <body>

        <h1>My First Heading</h1>
        <p>My first paragraph.</p>
        <p>My second paragraph.</p>
        </body>
        </html>
    """


def test_text_cleaning_with_sentence_segmentation():
    tokenizer = Tokenizer(
        text_cleaning=DefaultTextCleaning(rules=["html_to_text", "strip_spaces"]),
        segment_sentences=True,
    )

    tokenized = tokenizer.tokenize_text(html_text)
    assert len(tokenized) == 2
    assert (
        len(tokenized[0]) == 7
    ), "Expected [My, First, Heading, My, first, paragraph, .]"
    assert len(tokenized[1]) == 4, "Expected [My, second, paragraph, .]"


def test_text_cleaning_with_sentence_segmentation_and_max_sequence():
    tokenizer = Tokenizer(
        max_sequence_length=8,
        text_cleaning=DefaultTextCleaning(rules=["html_to_text", "strip_spaces"]),
        segment_sentences=True,
    )

    tokenized = tokenizer.tokenize_text(html_text)
    assert len(tokenized) == 2
    assert len(tokenized[0]) == 2, "Expected [My, First]"
    assert len(tokenized[1]) == 2, "Expected [My, second]"


def test_document_cleaning():
    tokenizer = Tokenizer(
        text_cleaning=DefaultTextCleaning(rules=["html_to_text", "strip_spaces"]),
        segment_sentences=True,
    )

    tokenized = tokenizer.tokenize_document([html_text])
    assert len(tokenized) == 2
    assert (
        len(tokenized[0]) == 7
    ), "Expected [My, First, Heading, My, first, paragraph, .]"
    assert len(tokenized[1]) == 4, "Expected [My, second, paragraph, .]"



from typing import Any, Dict, List, Union

from allennlp.data import Instance, Token, TokenIndexer
from allennlp.data.fields import ListField, TextField
from biome.text.tokenizer import Tokenizer
from biome.text.features import WordFeatures


class InputFeaturizer:
    """Transforms input text (words and/or characters) into indexes and embedding vectors.

    This class defines two input features, words and chars for embeddings at word and character level respectively.

    You can provide additional features by manually specify `indexer` and `embedder` configurations within each
    input feature.

    Attributes
    ----------
    tokenizer : `Tokenizer`
        Tokenizes the input depending on its type (str, List[str], Dict[str, Any])
    indexer : `Dict[str, TokenIdexer]`
        Features dictionary for token indexing. Built from `FeaturesConfiguration`
    """

    def __init__(
        self, tokenizer: Tokenizer, indexer: Dict[str, TokenIndexer],
    ):
        self.tokenizer = tokenizer
        self.indexer = indexer

    @property
    def has_word_features(self) -> bool:
        """Checks if word features are already configured as part of the featurization"""
        return WordFeatures.namespace in self.indexer

    def featurize(
        self,
        record: Union[str, List[str], Dict[str, Any]],
        to_field: str = "record",
        aggregate: bool = False,
        tokenize: bool = True,
        exclude_record_keys: bool = False,
    ) -> Instance:
        """Generates an allennlp instance from a record input.

        Parameters
        ----------
        record: `Union[str, List[str], Dict[str, Any]]`
            Input data
        to_field: `str`
            The field name in the returned instance
        aggregate: `bool`
            If true, the returned instance will contain a single `TextField` with all record fields;
            If false, the instance will contain a `ListField` of `TextField`s.
        tokenize: `bool`
            If false, skip tokenization phase and pass record data as tokenized token list.
        exclude_record_keys: `bool`
            If true, excludes record keys from the output tokens in the featurization of dictionaries

        Returns
        -------
        instance: `Instance`
        """
        return self(record, to_field, aggregate, tokenize, exclude_record_keys)

    def __call__(
        self,
        record: Union[str, List[str], Dict[str, Any]],
        to_field: str = "record",
        aggregate: bool = False,
        tokenize: bool = True,
        exclude_record_keys: bool = False,
    ):
        """See `self.featurize()`"""
        # TODO: Allow exclude record keys in data tokenization phase
        data = record

        record_tokens = (
            self._data_tokens(data, exclude_record_keys)
            if tokenize
            else [[Token(t) for t in data]]
        )
        return Instance({to_field: self._tokens_to_field(record_tokens, aggregate)})

    def _data_tokens(
        self, data: Union[str, List[str], Dict[str, Any]], exclude_record_keys: bool
    ) -> List[List[Token]]:
        """Converts the input data into a list of list of tokens depending on its type.

        Parameters
        ----------
        data: Union[str, List[str], Dict[str, Any]]
            The input data
        exclude_record_keys: `bool`
            If true, excludes record keys from the output tokens in the featurization of dictionaries

        Returns
        -------
        list_of_list_of_tokens
        """
        if isinstance(data, dict):
            return self.tokenizer.tokenize_record(data, exclude_record_keys)
        if isinstance(data, str):
            return self.tokenizer.tokenize_text(data)
        return self.tokenizer.tokenize_document(data)

    def _tokens_to_field(
        self, tokens: List[List[Token]], aggregate: bool
    ) -> Union[ListField, TextField]:
        """Wraps the tokens with the indexer in one or several `TextField`s

        Parameters
        ----------
        tokens: List[List[Token]]
            The list of list of input tokens
        aggregate: `bool`
            If true, the returned instance will contain a single `TextField` with all record fields;
            If false, the instance will contain a `ListField` of `TextField`s.

        Returns
        -------
        field
            `TextField` if aggregate is true, `ListField` otherwise
        """
        if aggregate:
            return TextField(
                [token for entry_tokens in tokens for token in entry_tokens],
                self.indexer,
            )
        return ListField(
            [TextField(entry_tokens, self.indexer) for entry_tokens in tokens]
        )



import unittest

import dask
import dask.threaded


class DaskSupportTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        dask.config.set(scheduler="single-threaded")

    @classmethod
    def tearDownClass(cls):
        pass



from typing import Any, Dict, Iterable, List, Optional, Tuple, Union

from allennlp.common import FromParams
from allennlp.data import Token
from allennlp.data.tokenizers import SentenceSplitter, SpacyTokenizer
from allennlp.data.tokenizers.sentence_splitter import SpacySentenceSplitter

from biome.text.text_cleaning import DefaultTextCleaning, TextCleaning


class Tokenizer(FromParams):
    """Pre-processes and tokenizes the input text
    
    Transforms inputs (e.g., a text, a list of texts, etc.) into structures containing `allennlp.data.Token` objects.
    
    Use its arguments to configure the first stage of the pipeline (i.e., pre-processing a given set of text inputs.)
    
    Use methods for tokenization depending on the shape of the inputs
    (e.g., records with multiple fields, sentences lists).
    
    Parameters
    ----------
    lang
        The [spaCy model used](https://spacy.io/api/tokenizer) for tokenization is language dependent.
        For optimal performance, specify the language of your input data (default: "en").
    max_sequence_length: `int`
        Maximum length in characters for input texts truncated with `[:max_sequence_length]` after `TextCleaning`.
    max_nr_of_sentences: `int`
        Maximum number of sentences to keep when using `segment_sentences` truncated with `[:max_sequence_length]`.
    text_cleaning: `Optional[TextCleaning]`
        A `TextCleaning` configuration with pre-processing rules for cleaning up and transforming raw input text.
    segment_sentences:  `Union[bool, SentenceSplitter]`
        Whether to segment input texts in to sentences using the default `SentenceSplitter` or a given splitter.
    start_tokens: `Optional[List[str]]`
        A list of token strings to the sequence before tokenized input text.
    end_tokens: `Optional[List[str]]`
        A list of token strings to the sequence after tokenized input text.
    """

    # TODO: Check if it is possible to pass in a `TokenizerConfiguration` to avoid duplication of arguments

    def __init__(
        self,
        lang: str = "en",
        max_sequence_length: int = None,
        max_nr_of_sentences: int = None,
        text_cleaning: Optional[TextCleaning] = None,
        segment_sentences: Union[bool, SentenceSplitter] = False,
        start_tokens: Optional[List[str]] = None,
        end_tokens: Optional[List[str]] = None,
    ):
        self._fetch_spacy_model(lang)

        if segment_sentences is True:
            # TODO: check rule-based feat.
            segment_sentences = SpacySentenceSplitter(language=lang, rule_based=True)

        self.lang = lang
        self.segment_sentences = segment_sentences
        self.max_nr_of_sentences = max_nr_of_sentences
        self.max_sequence_length = max_sequence_length
        self.text_cleaning = text_cleaning or DefaultTextCleaning()

        self._base_tokenizer = SpacyTokenizer(
            language=self.lang, start_tokens=start_tokens, end_tokens=end_tokens,
        )

    def _fetch_spacy_model(self, lang):
        # Allennlp get_spacy_model method works only for fully named models (en_core_web_sm) but no
        # for already linked named (en, es)
        # This is a workaround for mitigate those kind of errors. Just loading one more time, it's ok.
        # See https://github.com/allenai/allennlp/issues/4201
        import spacy

        try:
            spacy.load(lang, disable=["vectors", "textcat", "tagger" "parser" "ner"])
        except OSError:
            spacy.cli.download(lang)

    def tokenize_text(self, text: str) -> List[List[Token]]:
        """
        Tokenizes a text string applying sentence segmentation, if enabled

        Parameters
        ----------
        text: `str`
            The input text

        Returns
        -------
        A list of list of `Token`.

        If no sentence segmentation is enabled, or just one sentence is found in text
        the first level list will contain just one element: the tokenized text.

        """
        return self.tokenize_document([text])

    def _tokenize(self, text: str) -> List[Token]:
        """Tokenizes an input text string

        The simplest case where your input is just a `str`. For a text tokenization with
        sentence segmentation, see `tokenize_text`


        Parameters
        ----------
        text: `str`
            The input text
        Returns
        -------
        tokens: `List[Token]`

        """
        return self._base_tokenizer.tokenize(text[: self.max_sequence_length])

    def tokenize_document(self, document: List[str]) -> List[List[Token]]:
        """ Tokenizes a document-like structure containing lists of text inputs

        Use this to account for hierarchical text structures (e.g., a paragraph is a list of sentences)

        Parameters
        ----------
        document: `List[str]`
            A `List` with text inputs, e.g., sentences
            
        Returns
        -------
        tokens: `List[List[Token]]`
        """
        sentences = [self.text_cleaning(text) for text in document]
        if self.segment_sentences:
            sentences = [
                sentence
                for sentences in self.segment_sentences.batch_split_sentences(sentences)
                for sentence in sentences
            ]
        return [
            self._tokenize(sentence)
            for sentence in sentences[: self.max_nr_of_sentences]
        ]

    def tokenize_record(
        self, record: Dict[str, Any], exclude_record_keys: bool
    ) -> List[List[Token]]:
        """ Tokenizes a record-like structure containing text inputs
        
        Use this to keep information about the record-like data structure as input features to the model.
        
        Parameters
        ----------
        record: `Dict[str, Any]`
            A `Dict` with arbitrary "fields" containing text.
        exclude_record_keys: `bool`
            If enabled, exclude tokens related to record key text
            
        Returns
        -------
        tokens: `List[List[Token]]`
            A list of tokenized fields as token list
        """
        data = self._sanitize_dict(record)
        if exclude_record_keys:
            return [
                sentence
                for key, value in data.items()
                for sentence in self.tokenize_text(value)
            ]

        return [
            tokenized_key + sentence
            for key, value in data.items()
            for tokenized_key in [self._tokenize(key)]
            for sentence in self.tokenize_text(value)
        ]

    def _value_as_string(self, value: Any) -> str:
        """Converts a value data into its string representation"""
        if isinstance(value, str):
            return value
        if isinstance(value, dict):
            return self._value_as_string(value.values())
        if isinstance(value, Iterable):
            return " ".join(map(self._value_as_string, value))
        return str(value)

    def _sanitize_dict(self, data: Dict[str, Any]) -> Dict[str, str]:
        """Transforms an data dictionary into a dictionary of strings values"""
        return {key: self._value_as_string(value) for key, value in data.items()}



import torch
from allennlp.training.metrics import Metric


class MultiLabelF1Measure(Metric):
    """
    Computes overall F1 for multilabel classification tasks.
    Predictions sent to the __call__ function are logits and it turns them into 0 or 1s.
    Used for `classification heads` with the `multilabel` parameter enabled.
    """

    def __init__(self) -> None:
        self._tp = 0.0
        self._fp = 0.0
        self._fn = 0.0

    def __call__(
        self, predictions: torch.LongTensor, gold_labels: torch.LongTensor, **kwargs
    ):
        """
        Parameters
        ----------
        predictions : ``torch.Tensor``, required.
            A tensor with logits of shape (batch_size, ..., num_labels).
        gold_labels : ``torch.Tensor``, required.
            A tensor of 0 and 1 predictions of shape (batch_size, ..., num_labels).
            :param **kwargs:
        """
        # turn logits into one-hot predictions
        predictions = (predictions.data > 0.0).long()
        self._tp += (predictions * gold_labels).sum().item()
        self._fp += (predictions * (1 - gold_labels)).sum().item()
        self._fn += ((1 - predictions) * gold_labels).sum().item()

    def get_metric(self, reset: bool = False):
        """
        Returns
        -------
        `Dict[str, float]`
            A Dict with:
            - precision : `float`
            - recall : `float`
            - f1-measure : `float`
        """
        predicted_positives = self._tp + self._fp
        actual_positives = self._tp + self._fn

        precision = self._tp / predicted_positives if predicted_positives > 0 else 0
        recall = self._tp / actual_positives if actual_positives > 0 else 0

        f1 = (
            2 * precision * recall / (precision + recall)
            if precision + recall > 0
            else 0
        )

        if reset:
            self.reset()

        return {"precision": precision, "recall": recall, "fscore": f1}

    def reset(self):
        self._tp = 0.0
        self._fp = 0.0
        self._fn = 0.0



import pytest
from allennlp.common import Params
from allennlp.common.checks import ConfigurationError

from biome.text.configuration import FeaturesConfiguration


def test_non_configurable_features():
    wrong_config = dict(ner=dict(embedding=15))
    with pytest.raises(TypeError):
        FeaturesConfiguration(**wrong_config)

    with pytest.raises(ConfigurationError):
        FeaturesConfiguration.from_params(Params(wrong_config))



import os

import pytest
from dask.dataframe import DataFrame

from biome.text.data.readers import ElasticsearchDataFrameReader
from tests.test_support import DaskSupportTest

ES_HOST = os.getenv("ES_HOST", "http://localhost:9200")
ES_INDEX = os.getenv("ES_INDEX", "test-index")
ES_DOC = os.getenv("ES_DOC", "_doc")


class ElasticsearchReaderTest(DaskSupportTest):
    @staticmethod
    def _load_data_to_elasticsearch(data, host: str, index: str, doc: str):
        from elasticsearch import Elasticsearch
        from elasticsearch import helpers

        client = Elasticsearch(hosts=host, http_compress=True)
        client.indices.delete(index, ignore_unavailable=True)

        def _generator(data):
            for document in data:
                yield {"_index": index, "_type": doc, "_source": document}

        helpers.bulk(client, _generator(data))
        del client

    @pytest.mark.skip(reason="no way of currently testing this")
    def test_load_data(self):

        self._load_data_to_elasticsearch(
            [dict(a=i, b=f"this is {i}") for i in range(1, 5000)],
            host=ES_HOST,
            index=ES_INDEX,
            doc=ES_DOC,
        )

        es_index = ElasticsearchDataFrameReader.read(
            source=ElasticsearchDataFrameReader.SOURCE_TYPE,
            es_host=ES_HOST,
            index=ES_INDEX,
            doc_type=ES_DOC,
        )

        self.assertTrue(
            isinstance(es_index, DataFrame),
            f"elasticsearch datasource is not a dataframe :{type(es_index)}",
        )
        print(es_index.columns)
        self.assertTrue(
            "_id" not in es_index.columns and "id" in es_index.columns,
            "Expected renamed elasticsearch id",
        )



import click
from click import Path

from biome.text import Pipeline


@click.command("serve", help="Serves pipeline as rest api service")
@click.argument("pipeline_path", type=Path(exists=True))
@click.option("-p", "--port", "port", type=int, default=8888, show_default=True)
@click.option(
    "--predictions-dir",
    "predictions_dir",
    type=str,
    default=None,
    show_default=True,
    help="Path to log raw predictions from the service.",
)
def serve(pipeline_path: str, port: int, predictions_dir: str) -> None:
    pipeline = Pipeline.from_pretrained(pipeline_path)

    if predictions_dir:
        pipeline.init_prediction_logger(predictions_dir)

    pipeline.serve(port)



import os
import sys

import click
from click import Group

from .explore import explore
from .serve import serve
from .train import learn

SUPPORTED_COMMANDS = [explore, serve, learn]


def main():
    _add_project_modules_to_sys_path()

    commands = Group(no_args_is_help=True)
    for command in SUPPORTED_COMMANDS:
        commands.add_command(command, command.name)
    click.CommandCollection(sources=[commands])()


def _add_project_modules_to_sys_path():
    """This methods allows load udf defined from project location"""
    sys.path.append(os.getcwd())



import logging
from glob import glob
from typing import List, Optional, Union
from urllib.parse import urlparse

import dask.dataframe as dd
import pandas as pd
from dask import delayed
from dask_elk.client import DaskElasticClient

from .helpers import flatten_dask_dataframe, flatten_dataframe

ID = "id"
RESOURCE = "resource"
PATH_COLUMN_NAME = "path"

__LOGGER = logging.getLogger(__name__)
# TODO: The idea is to make the readers a class and define a metaclass that they have to follow.
#       For now, all reader methods have to return a dask.DataFrame. See ElasticsearchDataFrameReader


class DataFrameReader:
    """A base class for read :class:dask.dataframe.DataFrame
    """

    @classmethod
    def read(cls, source: Union[str, List[str]], **kwargs) -> dd.DataFrame:
        """
        Base class method for read the DataSources as a :class:dask.dataframe.DataFrame

        Parameters
        ----------
        source: The source information.
        kwargs: extra arguments passed to read method. Each reader should declare needed arguments

        Returns
        -------

        A :class:dask.dataframe.DataFrame read from source

        """

        raise NotImplementedError


def from_csv(path: Union[str, List[str]], **params) -> dd.DataFrame:
    """Creates a `dask.dataframe.DataFrame` from one or several csv files.
    Includes a "path column".

    Parameters
    ----------
    path
        Path to files
    params
        Extra arguments passed on to `dask.dataframe.read_csv`

    Returns
    -------
    df
        A `dask.DataFrame`

    """
    return dd.read_csv(path, include_path_column=PATH_COLUMN_NAME, **params)


def from_json(
    path: Union[str, List[str]], flatten: bool = True, **params
) -> dd.DataFrame:
    """Creates a `dask.dataframe.DataFrame` from one or several json files.
    Includes a "path column".

    Parameters
    ----------
    path
        Path to files
    flatten
        If true (default false), flatten json nested data
    params
        Extra arguments passed on to `pandas.read_json`

    Returns
    -------
    df
        A `dask.DataFrame`
    """

    def json_engine(*args, flatten: bool = False, **kwargs) -> pd.DataFrame:
        data_frame = pd.read_json(*args, **kwargs)
        return flatten_dataframe(data_frame) if flatten else data_frame

    path_list = _get_file_paths(path)

    dds = []
    for path_name in path_list:
        ddf = dd.read_json(path_name, flatten=flatten, engine=json_engine, **params)
        ddf[PATH_COLUMN_NAME] = path_name
        dds.append(ddf)

    return dd.concat(dds)


def from_parquet(path: Union[str, List[str]], **params) -> dd.DataFrame:
    """Creates a `dask.dataframe.DataFrame` from one or several parquet files.
    Includes a "path column".

    Parameters
    ----------
    path
        Path to files
    params
        Extra arguments passed on to `pandas.read_parquet`

    Returns
    -------
    df
        A `dask.dataframe.DataFrame`
    """
    path_list = _get_file_paths(path)

    dds = []
    for path_name in path_list:
        ddf = dd.read_parquet(path_name, engine="pyarrow", **params)
        ddf[PATH_COLUMN_NAME] = path_name
        dds.append(ddf)

    return dd.concat(dds)


def from_excel(path: Union[str, List[str]], **params) -> dd.DataFrame:
    """Creates a `dask.dataframe.DataFrame` from one or several excel files.
    Includes a "path column".

    Parameters
    ----------
    path
        Path to files
    params
        Extra arguments passed on to `pandas.read_excel`

    Returns
    -------
    df
        A `dask.dataframe.DataFrame`
    """
    path_list = _get_file_paths(path)

    dds = []
    for path_name in path_list:
        parts = delayed(pd.read_excel)(path_name, **params)
        data_frame = dd.from_delayed(parts).fillna("")
        data_frame[PATH_COLUMN_NAME] = path_name
        dds.append(data_frame)

    return dd.concat(dds)


def _get_file_paths(paths: Union[str, List[str]]) -> List[str]:
    """Return a list of path names that match the path names in paths.
    The path names can contain shell-style wildcards.

    Parameters
    ----------
    paths
        A path name or a list of path names. These path names can contain wildcards.

    Returns
    -------
    list_of_paths
        A list of path names.
    """
    if isinstance(paths, str):
        path_list = glob(paths)
        return path_list if path_list else [paths]
    path_lists = [_get_file_paths(path) for path in paths]

    # flatten the list of lists
    return [path for sublist in path_lists for path in sublist]


class ElasticsearchDataFrameReader(DataFrameReader):
    """
        Read a :class:dask.dataframe.DataFrame from a elasticsearch index
    """

    SOURCE_TYPE = "elasticsearch"
    __ELASTIC_ID_FIELD = "_id"

    @classmethod
    def read(  # pylint: disable=arguments-differ
        cls,
        source: Union[str, List[str]],
        index: str,
        doc_type: str = "_doc",
        query: Optional[dict] = None,
        es_host: str = "http://localhost:9200",
        flatten_content: bool = False,
        **kwargs,
    ) -> dd.DataFrame:
        """
        Creates a :class:dask.dataframe.DataFrame from a elasticsearch indexes

        Parameters
        ----------
        source
            The source param must match with :class:ElasticsearchDataFrameReader.SOURCE_TYPE
        es_host
            The elasticsearch host url (default to "http://localhost:9200")
        index
            The elasticsearch index
        doc_type
            The elasticsearch document type (default to "_doc")
        query
            The index query applied for extract the data
        flatten_content
            If True, applies a flatten to all nested data. It may take time to apply this flatten, so
            is deactivate by default.
        kwargs
            Extra arguments passed to base search method

        Returns
        -------
        A :class:dask.dataframe.DataFrame with index query results

        """
        if source != cls.SOURCE_TYPE:
            raise TypeError(f"{source} is not processable with {cls.__class__} reader")

        url = urlparse(es_host)
        client = DaskElasticClient(host=url.hostname, port=url.port, wan_only=True)
        data_frame = client.read(
            query=query, index=index, doc_type=doc_type, **kwargs
        ).persist()

        if flatten_content:
            data_frame = flatten_dask_dataframe(data_frame)
        data_frame[RESOURCE] = f"{index}/{doc_type}"

        if cls.__ELASTIC_ID_FIELD in data_frame.columns:
            data_frame = data_frame.rename(columns={cls.__ELASTIC_ID_FIELD: ID})

        return data_frame



from datetime import datetime

__version__ = "1.0.0rc2"

from typing import Optional


def package_version(version: str):
    """Generates a proper package version from git repository info"""

    def get_commit_hash(repository: git.Git) -> str:
        """
        Fetch current commit hash from configured git repository. The working folder should
        be part of a already git repository

        """
        return repository.log("--pretty=format:'%h'", "-n 1").replace("'", "")

    def get_first_tag_for_commit(repository: git.Git, commit_hash: str) -> str:
        """ Return tags related to current commit """

        tags = (
            repository.tag("--contains", commit_hash)
            if commit_hash
            else repository.tag("--contains")
        ).split("\n")
        return tags[0]

    def get_version_from_branch(repository: git.Git) -> Optional[str]:
        """Extract version from release branch name"""
        current_remote_branch = repository.branch("--remote", "--contains")
        if "releases/" in current_remote_branch:
            return current_remote_branch.split("/")[-1]
        return None

    def version_matches(release_branch_version: str, configured_version: str) -> bool:
        """Checks if a version matches with related release branch version"""
        minor_release = tuple(release_branch_version.split(".")[:2])
        minor_configured = tuple(configured_version.split(".")[:2])

        return minor_release == minor_configured

    try:
        repo = git.Git()
    except Exception:  # pylint: disable=broad-except
        return version

    commit = get_commit_hash(repo)
    repo_tag = get_first_tag_for_commit(repo, commit)
    release_version = get_version_from_branch(repo)
    if release_version and not repo_tag:
        assert version_matches(release_version, version)
        return version
    today = datetime.today().strftime("%Y%m%d%H%M%S")

    return repo_tag if repo_tag else f"{version}.dev{today}+{commit}"


try:
    import git

    __version__ = package_version(__version__)
# pylint: disable=broad-except
except Exception:
    pass



import json
import os
import tempfile
from typing import Dict

import logging

logging.basicConfig(level=logging.DEBUG)

TEST_RESOURCES = os.path.dirname(__file__)

os.chdir(TEST_RESOURCES)


def create_temp_configuration(data: Dict) -> str:
    with tempfile.NamedTemporaryFile(mode="w+", delete=False) as config_file:
        config_file.writelines(json.dumps(data).split("\n"))
        config_file.close()
        return config_file.name



import re
from pathlib import Path

from pytest_notebook.nb_regression import NBRegressionFixture
from pytest_notebook.notebook import load_notebook, dump_notebook

from tests import TUTORIALS_PATH


def test_text_classifier_tutorial(tmp_path):
    notebook_path = Path(TUTORIALS_PATH) / "Training_a_text_classifier.ipynb"

    # adapt notebook to CI (make its execution quicker + comment lines)
    notebook = load_notebook(str(notebook_path))
    for cell in notebook["cells"]:
        if cell["source"].startswith("!pip install"):
            cell["source"] = re.sub(r"!pip install", r"#!pip install", cell["source"])
        if cell["source"].startswith("trainer_config ="):
            cell["source"] = re.sub(
                r"num_epochs=[0-9][0-9]?", r"num_epochs=1", cell["source"]
            )
        if cell["source"].startswith("pl.train("):
            cell["source"] = re.sub(
                r"training=train_ds", r"training=valid_ds", cell["source"]
            )
        if cell["source"].startswith("pl_trained.explore"):
            cell["source"] = re.sub(
                r"pl_trained.explore", r"#pl_trained.explore", cell["source"]
            )

    # dump adapted notebook
    mod_notebook_path = tmp_path / notebook_path.name
    with mod_notebook_path.open("w") as file:
        file.write(str(dump_notebook(notebook)))

    # test adapted notebook
    fixture = NBRegressionFixture(exec_timeout=100)
    fixture.check(str(mod_notebook_path))



import os

from biome.text.data import DataSource
from tests import DaskSupportTest, RESOURCES_PATH

FILES_PATH = os.path.join(RESOURCES_PATH, "data")


class ParquetDataSourceTest(DaskSupportTest):
    def test_read_parquet(self):
        file_path = os.path.join(FILES_PATH, "test.parquet")
        ds = DataSource(format="parquet", source=file_path)

        df = ds.to_dataframe().compute()
        self.assertTrue("reviewerID" in df.columns)
        self.assertTrue("path" in df.columns)



from fastapi import HTTPException


class BaseError(Exception):
    """Base error. This class could include common error attributes or methods"""

    pass


class ValidationError(BaseError):
    """Base error for data validation"""

    pass


class MissingArgumentError(ValidationError):
    """Error related with input params"""

    def __init__(self, arg_name: str):
        super(MissingArgumentError, self).__init__()
        self.arg_name = arg_name

    def __str__(self) -> str:
        return f"Missing argument '{self.arg_name}'"


class ActionNotSupportedError(ValidationError):
    """Raised when an action is not supported for a given component state"""


class EmptyVocabError(ValidationError):
    """Error related with using empty vocabs for a training"""

    pass


class WrongValueError(ValidationError):
    """Wrong value error"""


class http_error_handling:
    """Error handling for http error transcription"""

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        if isinstance(exc_val, ValidationError):
            raise HTTPException(status_code=400, detail=str(exc_val))
        if isinstance(exc_val, Exception):
            # Common http error handling
            raise HTTPException(status_code=500, detail=str(exc_val))



from typing import Any, Dict, List, Optional, Union

import numpy
import torch
from allennlp.data import Instance
from allennlp.data.fields import LabelField, MultiLabelField
from allennlp.training.metrics import CategoricalAccuracy, FBetaMeasure

from biome.text import helpers, vocabulary
from biome.text.backbone import ModelBackbone
from biome.text.metrics import MultiLabelF1Measure
from ..task_head import TaskHead, TaskName, TaskOutput


class ClassificationHead(TaskHead):
    """Base abstract class for classification problems"""

    def __init__(
        self, backbone: ModelBackbone, labels: List[str], multilabel: bool = False
    ):
        super(ClassificationHead, self).__init__(backbone)
        vocabulary.set_labels(self.backbone.vocab, labels)

        # label related configurations
        self._multilabel = multilabel
        self.calculate_output = (
            self.multi_label_output if self._multilabel else self.single_label_output
        )

        # metrics and loss
        if self._multilabel:
            self.metrics = {"macro": MultiLabelF1Measure()}
            self._loss = torch.nn.BCEWithLogitsLoss()
        else:
            self.metrics = {"accuracy": CategoricalAccuracy()}
            self.metrics.update(
                {
                    "micro": FBetaMeasure(average="micro"),
                    "macro": FBetaMeasure(average="macro"),
                    "per_label": FBetaMeasure(
                        labels=[i for i in range(0, len(labels))]
                    ),
                }
            )
            self._loss = torch.nn.CrossEntropyLoss()

    def forward(self, *args: Any, **kwargs: Any) -> TaskOutput:
        raise NotImplementedError

    def featurize(self, *args, **kwargs) -> Optional[Instance]:
        raise NotImplementedError

    def add_label(
        self,
        instance: Instance,
        label: Union[List[str], List[int], str, int],
        to_field: str = "label",
    ) -> Optional[Instance]:
        """Includes the label field for classification into the instance data"""
        # "if not label:" fails for ndarrays this is why we explicitly check None
        if label is None:
            return instance

        field = None
        if (
            isinstance(label, numpy.ndarray) or isinstance(label, list)
        ) and self._multilabel:
            label = label.tolist() if isinstance(label, numpy.ndarray) else label
            field = MultiLabelField(label, label_namespace=vocabulary.LABELS_NAMESPACE)
        if isinstance(label, (str, int)) and not self._multilabel:
            field = LabelField(label, label_namespace=vocabulary.LABELS_NAMESPACE)
        if (
            not field
        ):  # We have label info but we cannot build the label field --> discard the instance
            return None

        instance.add_field(to_field, field)
        return instance

    def task_name(self) -> TaskName:
        return TaskName.text_classification

    def decode(self, output: TaskOutput) -> TaskOutput:
        """Completes the output for the prediction

        Mainly adds probabilities and keys for the UI.

        Parameters
        ----------
        output
            The output from the head's forward method

        Returns
        -------
        completed_output
        """
        if self._multilabel:
            probabilities = output.logits.sigmoid()
        else:
            probabilities = torch.nn.functional.softmax(output.logits, dim=-1)
        output.probs = probabilities

        output_map_probs = []
        max_classes = []
        max_classes_prob = []
        if self.num_labels > 0:
            for probs in probabilities:
                labels_with_prob = self._labels_with_probabilities(probs)
                output_map_probs.append(labels_with_prob)

                label, prob = list(labels_with_prob.items())[0]
                max_classes.append(label)
                max_classes_prob.append(prob)

        output.classes = output_map_probs

        if not self._multilabel:
            output.max_class = max_classes  # deprecated
            output.max_class_prob = max_classes_prob  # deprecated

            output.label = max_classes
            output.prob = max_classes_prob

        return output

    def _labels_with_probabilities(
        self, probabilities: torch.Tensor
    ) -> Dict[str, float]:
        """
        Calculates the descendant sorted label + probs dictionary
        using all output classes (not only predicted)
        """
        all_classes_probs = torch.zeros(
            self.num_labels,
            device=probabilities.get_device()
            if probabilities.get_device() > -1
            else None,
        )
        all_classes_probs[: probabilities.size()[0]] = probabilities
        sorted_indexes_by_prob = torch.argsort(
            all_classes_probs, descending=True
        ).tolist()

        return {
            vocabulary.label_for_index(self.backbone.vocab, idx): all_classes_probs[idx]
            for idx in sorted_indexes_by_prob
        }

    def get_metrics(self, reset: bool = False) -> Dict[str, float]:
        """Get the metrics of our classifier, see :func:`~allennlp_2.models.Model.get_metrics`.

        Parameters
        ----------
        reset
            Reset the metrics after obtaining them?

        Returns
        -------
        A dictionary with all metric names and values.
        """
        final_metrics = {}
        if "accuracy" in self.metrics.keys():
            final_metrics.update(
                {"accuracy": self.metrics["accuracy"].get_metric(reset)}
            )

        for metric_name in ["micro", "macro"]:
            if metric_name in self.metrics.keys():
                for k, v in self.metrics[metric_name].get_metric(reset).items():
                    final_metrics.update({"{}/{}".format(metric_name, k): v})

        if "per_label" in self.metrics.keys():
            for k, values in self.metrics["per_label"].get_metric(reset).items():
                for i, v in enumerate(values):
                    label = vocabulary.label_for_index(self.backbone.vocab, i)
                    # sanitize label using same patterns as tensorboardX to avoid summary writer warnings
                    label = helpers.clean_metric_name(label)
                    final_metrics.update({"_{}/{}".format(k, label): v})

        return final_metrics

    def single_label_output(
        self, logits: torch.Tensor, label: Optional[torch.IntTensor] = None,
    ) -> TaskOutput:
        output = TaskOutput(logits=logits)

        if label is not None:
            output.loss = self._loss(logits, label.long())
            for metric in self.metrics.values():
                metric(logits, label)

        return output

    def multi_label_output(
        self, logits: torch.Tensor, label: Optional[torch.IntTensor] = None,
    ) -> TaskOutput:
        output = TaskOutput(logits=logits)

        if label is not None:
            # casting long to float for BCELoss
            # see https://discuss.pytorch.org/t/nn-bcewithlogitsloss-cant-accept-one-hot-target/59980
            output.loss = self._loss(
                logits.view(-1, self.num_labels),
                label.view(-1, self.num_labels).type_as(logits),
            )
            for metric in self.metrics.values():
                metric(logits, label)

        return output



import copy
import inspect
from typing import Any, Dict, Generic, Optional, Type, TypeVar, Union

from allennlp.common import FromParams, Params
from allennlp.modules.bimpm_matching import BiMpmMatching
from allennlp.modules.seq2seq_encoders import PytorchSeq2SeqWrapper
from allennlp.modules.seq2vec_encoders import PytorchSeq2VecWrapper

from biome.text import helpers


def _find_input_attribute(component: Type[Any]) -> str:
    """Find the properly input dimension attribute name for a given component"""
    input_dim_attribute = None
    if issubclass(component, (PytorchSeq2SeqWrapper, PytorchSeq2VecWrapper)):
        input_dim_attribute = "input_size"
    elif component is BiMpmMatching:
        input_dim_attribute = "hidden_dim"
    else:
        init_method_keys = inspect.signature(component.__init__).parameters.keys()
        for param_name in ["embedding_dim", "input_dim"]:
            if param_name in init_method_keys:
                input_dim_attribute = param_name
                break
    return input_dim_attribute


T = TypeVar("T")


class ComponentSpec(Generic[T], FromParams):
    """
    The layer spec component allows create Pytorch modules lazily,
    and instantiate them inside a context (Model or other component) dimension layer chain.

    The layer spec wraps a component params and will generate an instance of type T once the input_dim is set.

    """

    @classmethod
    def from_params(cls: Type[T], params: Params, **extras) -> T:
        return cls(**params.as_dict())

    def __resolve_layer_class(
        self, type_name: Optional[Union[Type, str]] = None
    ) -> Type[T]:
        if isinstance(type_name, Type):
            return type_name

        layer_class = getattr(self.__class__, "__orig_bases__")[0].__args__[0]
        return layer_class.by_name(type_name) if type_name else layer_class

    def __init__(self, **config):
        self._layer_class = self.__resolve_layer_class(config.get("type"))
        config["type"] = helpers.get_full_class_name(self._layer_class)
        self._config = config or {}

    def input_dim(self, input_dim: int) -> "ComponentSpec":
        """Sets the input dimension attribute for this layer configuration"""
        self.__update_config_with_input_dim(input_dim)
        return self

    def __update_config_with_input_dim(self, input_dim: int):
        input_dim_attribute = _find_input_attribute(self._layer_class)

        if input_dim_attribute:
            self._config[input_dim_attribute] = input_dim

    @property
    def config(self) -> Dict[str, Any]:
        """Component read-only configuration"""
        return copy.deepcopy(self._config)

    def compile(self, **extras) -> T:
        """
        Using the wrapped configuration and the input dimension, generates a
        instance of type T representing the layer configuration
        """
        if not self.config:
            raise ValueError(f"No configuration found for {self}")

        config = self.config
        if "type" in config:
            config.pop("type")

        return self._layer_class.from_params(Params(config), **extras)



from typing import Any, Dict, List, Optional, Union, cast

import numpy
import torch
from allennlp.data import Batch, Instance, TextFieldTensors
from allennlp.data.fields import ListField, TextField
from allennlp.modules.seq2seq_encoders import PassThroughEncoder
from allennlp.modules.seq2vec_encoders import BagOfEmbeddingsEncoder
from allennlp.nn.util import get_text_field_mask
from captum.attr import IntegratedGradients

from biome.text import vocabulary
from biome.text.backbone import ModelBackbone
from biome.text.modules.encoders import TimeDistributedEncoder
from biome.text.modules.specs import (
    ComponentSpec,
    FeedForwardSpec,
    Seq2SeqEncoderSpec,
    Seq2VecEncoderSpec,
)
from .classification import ClassificationHead
from ..task_head import TaskOutput


class DocumentClassification(ClassificationHead):
    """
    Task head for document text classification. It's quite similar to text
    classification but including the doc2vec transformation layers
    """

    forward_arg_name = "document"
    label_name = "label"

    def __init__(
        self,
        backbone: ModelBackbone,
        labels: List[str],
        tokens_pooler: Optional[Seq2VecEncoderSpec] = None,
        sentences_encoder: Optional[Seq2SeqEncoderSpec] = None,
        sentences_pooler: Seq2VecEncoderSpec = None,
        feedforward: Optional[FeedForwardSpec] = None,
        multilabel: bool = False,
    ) -> None:

        super(DocumentClassification, self).__init__(
            backbone, labels=labels, multilabel=multilabel
        )

        self.backbone.encoder = TimeDistributedEncoder(backbone.encoder)

        # layers
        self.tokens_pooler = TimeDistributedEncoder(
            BagOfEmbeddingsEncoder(embedding_dim=self.backbone.encoder.get_output_dim())
            if not tokens_pooler
            else tokens_pooler.input_dim(
                self.backbone.encoder.get_output_dim()
            ).compile()
        )
        self.sentences_encoder = (
            PassThroughEncoder(self.tokens_pooler.get_output_dim())
            if not sentences_encoder
            else sentences_encoder.input_dim(
                self.tokens_pooler.get_output_dim()
            ).compile()
        )
        self.sentences_pooler = (
            BagOfEmbeddingsEncoder(self.sentences_encoder.get_output_dim())
            if not sentences_pooler
            else sentences_pooler.input_dim(
                self.sentences_encoder.get_output_dim()
            ).compile()
        )
        self.feedforward = (
            None
            if not feedforward
            else feedforward.input_dim(self.sentences_pooler.get_output_dim()).compile()
        )

        self._classification_layer = torch.nn.Linear(
            (self.feedforward or self.sentences_pooler).get_output_dim(),
            self.num_labels,
        )

    def featurize(
        self,
        document: Any,
        label: Optional[Union[int, str, List[Union[int, str]]]] = None,
    ) -> Optional[Instance]:
        instance = self.backbone.featurizer(
            document, to_field=self.forward_arg_name, exclude_record_keys=True
        )
        return self.add_label(instance, label, to_field=self.label_name)

    def forward(
        self, document: TextFieldTensors, label: torch.IntTensor = None
    ) -> TaskOutput:
        mask = get_text_field_mask(
            document, num_wrapping_dims=1
        )  # Why num_wrapping_dims=1 !?
        embedded_text = self.backbone.forward(document, mask, num_wrapping_dims=1)
        embedded_text = self.tokens_pooler(embedded_text, mask=mask)

        # Here we need to mask the TextFields that only contain the padding token -> last dimension only contains False
        # Those fields were added to possibly equalize the batch.
        mask = torch.sum(mask, -1) > 0
        embedded_text = self.sentences_encoder(embedded_text, mask=mask)
        embedded_text = self.sentences_pooler(embedded_text, mask=mask)

        if self.feedforward is not None:
            embedded_text = self.feedforward(embedded_text)

        logits = self._classification_layer(embedded_text)
        return self.calculate_output(logits=logits, label=label)

    def explain_prediction(
        self, prediction: Dict[str, numpy.array], instance: Instance, n_steps: int
    ) -> Dict[str, Any]:
        """Here, we must apply transformations for manage ListFields tensors shapes"""

        dataset = Batch([instance])
        input_tokens_ids = dataset.as_tensor_dict()
        ig = IntegratedGradients(self._explain_embeddings)

        num_wrapping_dims = 1

        document_tokens = [
            [token.text for token in cast(TextField, text_field).tokens]
            for text_field in cast(ListField, instance.get(self.forward_arg_name))
        ]
        document_tensors = input_tokens_ids.get(self.forward_arg_name)
        mask = get_text_field_mask(
            document_tensors, num_wrapping_dims=num_wrapping_dims
        )
        text_embeddings = self.backbone.embedder.forward(
            document_tensors, num_wrapping_dims=num_wrapping_dims
        )

        label_id = vocabulary.index_for_label(
            self.backbone.vocab, prediction.get(self.label_name)
        )
        attributions, delta = ig.attribute(
            text_embeddings,
            target=label_id,
            additional_forward_args=mask,
            return_convergence_delta=True,
            n_steps=n_steps,
        )
        # TODO: what the attribution and deltas means
        attributions = attributions.sum(dim=3).squeeze(0)
        attributions = attributions / torch.norm(attributions)
        attributions = attributions.detach().numpy()

        return {
            **prediction,
            "explain": {
                self.forward_arg_name: [
                    [
                        {"token": token, "attribution": attribution}
                        for token, attribution in zip(
                            sentence_tokens, sentence_attribution
                        )
                    ]
                    for sentence_tokens, sentence_attribution in zip(
                        document_tokens, attributions
                    )
                ]
            },
        }

    def _explain_embeddings(
        self, embeddings: torch.Tensor, mask: torch.Tensor
    ) -> torch.Tensor:
        """Embedding interpret for ListField shapes"""
        embedded_text = self.backbone.encoder.forward(embeddings, mask)
        embedded_text = self.tokens_pooler(embedded_text, mask=mask)

        sentences_mask = torch.sum(mask, -1) > 0
        embedded_text = self.sentences_encoder(embedded_text, mask=sentences_mask)
        embedded_text = self.sentences_pooler(embedded_text, mask=sentences_mask)

        if self.feedforward is not None:
            embedded_text = self.feedforward(embedded_text)

        logits = self._classification_layer(embedded_text)
        return logits


class DocumentClassificationSpec(ComponentSpec[DocumentClassification]):
    """Lazy initialization for document classification head components"""

    pass



import copy
from typing import Any, Dict, List, Optional, Type, Union

import yaml
from allennlp.common import FromParams, Params
from allennlp.data import TokenIndexer, Vocabulary
from allennlp.modules import TextFieldEmbedder

from biome.text.data import DataSource
from . import vocabulary
from .features import CharFeatures, WordFeatures
from .featurizer import InputFeaturizer
from .helpers import save_dict_as_yaml
from .modules.encoders import Encoder
from .modules.heads.task_head import TaskHeadSpec
from .tokenizer import Tokenizer


class FeaturesConfiguration(FromParams):
    """Configures the input features of the `Pipeline`

    Use this for defining the features to be used by the model, namely word and character embeddings.
    
    :::tip
    If you do not pass in either of the parameters (`word` or `char`),
    your pipeline will be setup with a default word feature (embedding_dim=50).
    :::
    
    Example:
    
    ```python
    word = WordFeatures(embedding_dim=100)
    char = CharFeatures(embedding_dim=16, encoder={'type': 'gru'})
    config = FeaturesConfiguration(word, char)
    ```
    
    Parameters
    ----------
    word : `biome.text.features.WordFeatures`
        The word feature configurations, see `WordFeatures`
    char: `biome.text.features.CharFeatures`
        The character feature configurations, see `CharFeatures`
    """

    __DEFAULT_CONFIG = WordFeatures(embedding_dim=50)

    def __init__(
        self, word: Optional[WordFeatures] = None, char: Optional[CharFeatures] = None,
    ):
        self.word = word or None
        self.char = char or None

        if not (word or char):
            self.word = self.__DEFAULT_CONFIG

    @classmethod
    def from_params(
        cls: Type["FeaturesConfiguration"], params: Params, **extras
    ) -> "FeaturesConfiguration":

        word = params.pop("word", params.pop("words", None))  # TODO: remove backward
        word = WordFeatures(**word.as_dict(quiet=True)) if word else None

        char = params.pop("char", params.pop("chars", None))  # TODO: remove backward
        char = CharFeatures(**char.as_dict(quiet=True)) if char else None

        params.assert_empty("FeaturesConfiguration")
        return cls(word=word, char=char)

    @property
    def keys(self) -> List[str]:
        """Gets the keys of the features"""
        return [key for key in vars(self)]

    def compile_embedder(self, vocab: Vocabulary) -> TextFieldEmbedder:
        """Creates the embedder based on the configured input features

        Parameters
        ----------
        vocab: `Vocabulary`
            The vocabulary for which to create the embedder

        Returns
        -------
        embedder
        """
        configuration = self._make_allennlp_config()
        if vocabulary.is_empty(vocab, namespaces=self.keys) and configuration.get(
            "word"
        ):
            # We simplify embedder configuration for better load an blank pipeline which create the vocab
            embedder_cfg = configuration["word"]["embedder"]
            embedder_cfg["embedding_dim"] = 1
            if "pretrained_file" in embedder_cfg:
                embedder_cfg["pretrained_file"] = None
        return TextFieldEmbedder.from_params(
            Params(
                {
                    "token_embedders": {
                        feature: config["embedder"]
                        for feature, config in configuration.items()
                    }
                }
            ),
            vocab=vocab,
        )

    def compile_featurizer(self, tokenizer: Tokenizer) -> InputFeaturizer:
        """Creates the featurizer based on the configured input features

        :::tip
        If you are creating configurations programmatically
        use this method to check that you provided a valid configuration.
        :::

        Parameters
        ----------
        tokenizer: `Tokenizer`
            Tokenizer used for this featurizer

        Returns
        -------
        featurizer: `InputFeaturizer`
            The configured `InputFeaturizer`
        """
        configuration = self._make_allennlp_config()

        indexer = {
            feature: TokenIndexer.from_params(Params(config["indexer"]))
            for feature, config in configuration.items()
        }
        return InputFeaturizer(tokenizer, indexer=indexer)

    def _make_allennlp_config(self) -> Dict[str, Any]:
        """Returns a configuration dict compatible with allennlp

        Returns
        -------
        config_dict
        """
        configuration = {
            spec.namespace: spec.config for spec in [self.word, self.char] if spec
        }

        return copy.deepcopy(configuration)


class TokenizerConfiguration(FromParams):
    """Configures the `Tokenizer`

    For a description of the parameters see `biome.text.tokenizer.Tokenizer`

    Parameters
    ----------
    lang
        The [spaCy model used](https://spacy.io/api/tokenizer) for tokenization is language dependent.
        For optimal performance, specify the language of your input data (default: "en").
    max_sequence_length
    max_nr_of_sentences
    text_cleaning
    segment_sentences
    """

    def __init__(
        self,
        lang: str = "en",
        max_sequence_length: int = None,
        max_nr_of_sentences: int = None,
        text_cleaning: Optional[Dict[str, Any]] = None,
        segment_sentences: Union[bool, Dict[str, Any]] = False,
    ):
        self.lang = lang
        self.max_sequence_length = max_sequence_length
        self.max_nr_of_sentences = max_nr_of_sentences
        self.text_cleaning = text_cleaning
        self.segment_sentences = segment_sentences

    def compile(self) -> Tokenizer:
        """Build tokenizer object from its configuration"""
        return Tokenizer.from_params(Params(copy.deepcopy(vars(self))))


class PipelineConfiguration(FromParams):
    """"Creates a `Pipeline` configuration

    Parameters
    ----------
    name: `str`
        The `name` for our pipeline
    features: `FeaturesConfiguration`
        The input `features` to be used by the model pipeline. We define this using a `FeaturesConfiguration` object.
    head: `TaskHeadSpec`
        The `head` for the task, e.g., a LanguageModelling task, using a `TaskHeadSpec` object.
    tokenizer: `TokenizerConfiguration`, optional
        The `tokenizer` defined with a `TokenizerConfiguration` object.
    encoder: `Seq2SeqEncoderSpec`
        The core text seq2seq `encoder` of our model using a `Seq2SeqEncoderSpec`
    """

    def __init__(
        self,
        name: str,
        features: FeaturesConfiguration,
        head: TaskHeadSpec,
        tokenizer: Optional[TokenizerConfiguration] = None,
        encoder: Optional[Encoder] = None,
    ):
        super(PipelineConfiguration, self).__init__()

        self.name = name
        self.tokenizer = tokenizer or TokenizerConfiguration()
        self.features = features
        self.encoder = encoder
        self.head = head

    @classmethod
    def from_yaml(cls, path: str) -> "PipelineConfiguration":
        """Creates a pipeline configuration from a config yaml file

        Parameters
        ----------
        path: `str`
            The path to a YAML configuration file

        Returns
        -------
        pipeline_configuration: `PipelineConfiguration`
        """
        with open(path) as yaml_file:
            config_dict = yaml.safe_load(yaml_file)

        return cls.from_dict(config_dict)

    @classmethod
    def from_dict(cls, config_dict: dict) -> "PipelineConfiguration":
        """Creates a pipeline configuration from a config dictionary

        Parameters
        ----------
        config_dict: `dict`
            A configuration dictionary

        Returns
        -------
        pipeline_configuration: `PipelineConfiguration`
        """
        return PipelineConfiguration.from_params(Params(config_dict))

    def as_dict(self) -> Dict[str, Any]:
        config = {
            "name": self.name,
            "tokenizer": vars(self.tokenizer),
            "features": vars(self.features),
            "head": self.head.config,
        }

        if self.encoder:
            config["encoder"] = self.encoder.config

        return config

    def to_yaml(self, path: str):
        """Saves the pipeline configuration to a yaml formatted file

        Parameters
        ----------
        path : str
            Path to the output file
        """
        config_dict = copy.deepcopy(self.as_dict())
        config_dict["features"]["word"] = (
            config_dict["features"]["word"].to_dict()
            if config_dict["features"]["word"] is not None
            else None
        )
        config_dict["features"]["char"] = (
            config_dict["features"]["char"].to_dict()
            if config_dict["features"]["char"] is not None
            else None
        )

        save_dict_as_yaml(config_dict, path)

    def build_tokenizer(self) -> Tokenizer:
        """Build the pipeline tokenizer"""
        return self.tokenizer.compile()

    def build_featurizer(self) -> InputFeaturizer:
        """Creates the pipeline featurizer"""
        return self.features.compile_featurizer(self.tokenizer.compile())

    def build_embedder(self, vocab: Vocabulary):
        """Build the pipeline embedder for aiven dictionary"""
        return self.features.compile_embedder(vocab)


class TrainerConfiguration:
    """ Creates a `TrainerConfiguration`

    Parameters
    ----------
    optimizer
    validation_metric
    patience
    num_epochs
    cuda_device
    grad_norm
    grad_clipping
    learning_rate_scheduler
    momentum_scheduler
    moving_average
    batch_size
    cache_instances
    in_memory_batches
    data_bucketing
    """

    def __init__(
        self,
        optimizer: Dict[str, Any] = None,
        validation_metric: str = "-loss",
        patience: Optional[int] = 2,
        num_epochs: int = 20,
        cuda_device: int = -1,
        grad_norm: Optional[float] = None,
        grad_clipping: Optional[float] = None,
        learning_rate_scheduler: Optional[Dict[str, Any]] = None,
        momentum_scheduler: Optional[Dict[str, Any]] = None,
        moving_average: Optional[Dict[str, Any]] = None,
        batch_size: Optional[int] = 16,
        cache_instances: bool = True,
        in_memory_batches: int = 2,
        data_bucketing: bool = True,
    ):
        self.optimizer = optimizer or {"type": "adam"}
        self.validation_metric = validation_metric
        self.patience = patience
        self.num_epochs = num_epochs
        self.cuda_device = cuda_device
        self.grad_norm = grad_norm
        self.grad_clipping = grad_clipping
        self.learning_rate_scheduler = learning_rate_scheduler
        self.momentum_scheduler = momentum_scheduler
        self.moving_average = moving_average

        # Data Iteration
        self.batch_size = batch_size
        self.data_bucketing = data_bucketing
        self.cache_instances = cache_instances
        self.in_memory_batches = in_memory_batches


class VocabularyConfiguration:
    """Configures a ``Vocabulary`` before it gets created from the data

    Use this to configure a Vocabulary using specific arguments from `allennlp.data.Vocabulary``

    See [AllenNLP Vocabulary docs](https://docs.allennlp.org/master/api/data/vocabulary/#vocabulary])

    Parameters
    ----------
    sources: `List[DataSource]`
        Datasource to be used for data creation
    min_count: `Dict[str, int]`, optional (default=None)
        Minimum number of appearances of a token to be included in the vocabulary.
        The key in the dictionary refers to the namespace of the input feature.
    max_vocab_size:  `Union[int, Dict[str, int]]`, optional (default=`None`)
        Maximum number of tokens in the vocabulary
    pretrained_files: `Optional[Dict[str, str]]`, optional
        Pretrained files with word vectors
    only_include_pretrained_words: `bool`, optional (default=False)
        Only include tokens present in pretrained_files
    tokens_to_add: `Dict[str, int]`, optional
        A list of tokens to add to the vocabulary, even if they are not present in the ``sources``
    min_pretrained_embeddings: ``Dict[str, int]``, optional
        Minimum number of lines to keep from pretrained_files, even for tokens not appearing in the sources.
    """

    def __init__(
        self,
        sources: List[DataSource],
        min_count: Dict[str, int] = None,
        max_vocab_size: Union[int, Dict[str, int]] = None,
        pretrained_files: Optional[Dict[str, str]] = None,
        only_include_pretrained_words: bool = False,
        tokens_to_add: Dict[str, List[str]] = None,
        min_pretrained_embeddings: Dict[str, int] = None,
    ):
        self.sources = sources
        self.pretrained_files = pretrained_files
        self.min_count = min_count
        self.max_vocab_size = max_vocab_size
        self.only_include_pretrained_words = only_include_pretrained_words
        self.tokens_to_add = tokens_to_add
        self.min_pretrained_embeddings = min_pretrained_embeddings



import warnings
from warnings import warn_explicit

import pkg_resources

from .pipeline import (
    Pipeline,
    PipelineConfiguration,
    TrainerConfiguration,
    VocabularyConfiguration,
)

warnings.showwarning = warn_explicit

try:
    __version__ = pkg_resources.get_distribution(__name__.replace(".", "-")).version
except pkg_resources.DistributionNotFound:
    # package is not installed
    pass



import copy
import inspect
import logging
import time
from threading import Thread
from typing import Any, Dict, List
from urllib.error import URLError
from urllib.parse import urlparse

import uvicorn
from allennlp.common.util import sanitize
from dask import dataframe as dd
from dask_elk.client import DaskElasticClient
from fastapi import FastAPI

from biome.text import Pipeline, TrainerConfiguration, helpers
from biome.text._configuration import (
    ElasticsearchExplore,
    ExploreConfiguration,
    TrainConfiguration,
    _ModelImpl,
)
from biome.text.constants import EXPLORE_APP_ENDPOINT
from biome.text.data import DataSource
from biome.text.errors import http_error_handling
from biome.text.modules.encoders import TimeDistributedEncoder
from biome.text.ui import launch_ui

_LOGGER = logging.getLogger(__name__)


def _serve(pipeline: Pipeline, port: int):
    """Serves an pipeline as rest api"""

    def make_app() -> FastAPI:
        app = FastAPI()

        @app.post("/predict")
        async def predict(inputs: Dict[str, Any]):
            with http_error_handling():
                return sanitize(pipeline.predict(**inputs))

        @app.post("/explain")
        async def explain(inputs: Dict[str, Any]):
            with http_error_handling():
                return sanitize(pipeline.explain(**inputs))

        @app.get("/_config")
        async def config():
            with http_error_handling():
                return pipeline.config.as_dict()

        @app.get("/_status")
        async def status():
            with http_error_handling():
                return {"ok": True}

        return app

    uvicorn.run(make_app(), host="0.0.0.0", port=port)


def _allennlp_configuration(
    pipeline: Pipeline, config: TrainConfiguration
) -> Dict[str, Any]:
    """Creates a allennlp configuration for pipeline train experiment configuration"""

    def trainer_configuration(trainer: TrainerConfiguration) -> Dict[str, Any]:
        """Creates trainer configuration dict"""
        __excluded_keys = [
            "data_bucketing",
            "batch_size",
            "cache_instances",
            "in_memory_batches",
        ]  # Data iteration attributes
        trainer_config = {
            k: v for k, v in vars(trainer).items() if k not in __excluded_keys
        }
        trainer_config.update(
            {
                "checkpointer": {"num_serialized_models_to_keep": 1},
                "tensorboard_writer": {"should_log_learning_rate": True},
            }
        )
        return trainer_config

    def iterator_configuration(
        pipeline: Pipeline, trainer: TrainerConfiguration
    ) -> Dict[str, Any]:
        """Creates a data iterator configuration"""

        def _forward_inputs() -> List[str]:
            """
            Calculate the required head.forward arguments. We use this method
            for automatically generate data iterator sorting keys
            """
            required, _ = helpers.split_signature_params_by_predicate(
                pipeline.head.forward, lambda p: p.default == inspect.Parameter.empty,
            )
            return [p.name for p in required] or [None]

        iterator_config = {
            "batch_size": trainer.batch_size,
            "max_instances_in_memory": max(
                trainer.batch_size * trainer.in_memory_batches, trainer.batch_size,
            ),
            "cache_instances": trainer.cache_instances,
            "type": "basic",
        }

        if trainer.data_bucketing:
            iterator_config.update(
                {
                    "sorting_keys": [
                        [
                            _forward_inputs()[0],
                            "list_num_tokens"
                            if isinstance(
                                pipeline.backbone.encoder, TimeDistributedEncoder
                            )
                            else "num_tokens",
                        ]
                    ],
                    "type": "bucket",
                }
            )

        return iterator_config

    base_config = {
        "config": pipeline.config.as_dict(),
        "type": _ModelImpl.__name__,
    }
    allennlp_config = {
        "trainer": trainer_configuration(config.trainer),
        "iterator": iterator_configuration(pipeline, config.trainer),
        "dataset_reader": base_config,
        "model": base_config,
        "train_data_path": config.training,
        "validation_data_path": config.validation,
        "test_data_path": config.test,
    }
    return copy.deepcopy({k: v for k, v in allennlp_config.items() if v})


def _explore(
    pipeline: Pipeline,
    data_source: DataSource,
    config: ExploreConfiguration,
    elasticsearch: ElasticsearchExplore,
) -> dd.DataFrame:
    """
    Executes a pipeline prediction over a datasource and register results int a elasticsearch index

    Parameters
    ----------
    pipeline
    data_source
    config
    elasticsearch

    Returns
    -------

    """
    if config.prediction_cache > 0:
        # TODO: do it
        pipeline.init_prediction_cache(config.prediction_cache)

    ddf_mapped = data_source.to_mapped_dataframe()
    # Stringify input data for better elasticsearch index mapping integration,
    # avoiding properties with multiple value types (string and long,...)
    for column in ddf_mapped.columns:
        ddf_mapped[column] = ddf_mapped[column].apply(helpers.stringify)

    # this only makes really sense when we have a predict_batch_json method implemented ...
    n_partitions = max(1, round(len(ddf_mapped) / config.batch_size))

    # a persist is necessary here, otherwise it fails for n_partitions == 1
    # the reason is that with only 1 partition we pass on a generator to predict_batch_json
    ddf_mapped = ddf_mapped.repartition(npartitions=n_partitions).persist()

    apply_func = pipeline.explain if config.explain else pipeline.predict
    ddf_mapped["annotation"] = ddf_mapped[pipeline.inputs].apply(
        lambda x: sanitize(apply_func(**x.to_dict())), axis=1, meta=(None, object)
    )

    ddf_source = (
        data_source.to_dataframe().repartition(npartitions=n_partitions).persist()
    )
    # Keep as metadata only non used values/columns
    ddf_source = ddf_source[
        [c for c in ddf_source.columns if c not in ddf_mapped.columns]
    ]
    ddf_mapped["metadata"] = ddf_source.map_partitions(
        lambda df: helpers.stringify(sanitize(df.to_dict(orient="records")))
    )

    # TODO @dcfidalgo we could calculate base metrics here (F1, recall & precision) using dataframe.
    #  And include as part of explore metadata
    #  Does it's simple???

    ddf = DaskElasticClient(
        host=elasticsearch.es_host, retry_on_timeout=True, http_compress=True
    ).save(ddf_mapped, index=elasticsearch.es_index, doc_type=elasticsearch.es_doc)

    elasticsearch.create_explore_data_index(force_delete=config.force_delete)
    elasticsearch.create_explore_data_record(
        {
            **(config.metadata or {}),
            "datasource": data_source.source,
            # TODO this should change when ui is normalized (action detail and action link naming)F
            "explore_name": elasticsearch.es_index,
            "model": pipeline.name,
            "columns": ddf.columns.values.tolist(),
            "metadata_columns": data_source.to_dataframe().columns.values.tolist(),
            "pipeline": pipeline.type_name,
            "output": pipeline.output,
            "inputs": pipeline.inputs,  # backward compatibility
            "signature": pipeline.inputs + [pipeline.output],
            "predict_signature": pipeline.inputs,
            "labels": pipeline.head.labels,
            "task": pipeline.head.task_name().as_string(),
        }
    )
    return ddf.persist()


def _show_explore(elasticsearch: ElasticsearchExplore) -> None:
    """Shows explore ui for data prediction exploration"""

    def is_service_up(url: str) -> bool:
        import urllib.request

        try:
            status_code = urllib.request.urlopen(url).getcode()
            return 200 <= status_code < 400
        except URLError:
            return False

    def launch_ui_app(ui_port: int) -> Thread:
        process = Thread(
            target=launch_ui,
            name="ui",
            kwargs=dict(es_host=elasticsearch.es_host, port=ui_port),
        )
        process.start()
        return process

    def show_notebook_explore(url: str):
        """Shows explore ui in a notebook cell"""
        from IPython.core.display import HTML, display

        iframe = f"<iframe src={url} width=100% height=840></iframe>"
        display(HTML(iframe))

    def show_browser_explore(url: str):
        """Shows explore ui in a web browser"""
        import webbrowser

        webbrowser.open(url)

    waiting_seconds = 1
    url = f"{EXPLORE_APP_ENDPOINT}/{elasticsearch.es_index}"

    if not is_service_up(url):
        port = urlparse(EXPLORE_APP_ENDPOINT).port
        if not port:
            _LOGGER.warning(
                "Cannot start explore application. "
                "Please, be sure you can reach %s from your browser "
                "or configure 'BIOME_EXPLORE_ENDPOINT' environment variable",
                url,
            )
            return
        launch_ui_app(port)
    time.sleep(waiting_seconds)
    _LOGGER.info("You can access to your data exploration from this url: %s", url)
    show_func = (
        show_notebook_explore
        if helpers.is_running_on_notebook()
        else show_browser_explore
    )
    show_func(url)



import re

import click
from click import Path

from biome.text import Pipeline
from biome.text.constants import DEFAULT_ES_HOST
from biome.text.data import DataSource
from biome.text.environment import ES_HOST


def _sanizite_index(index_name: str) -> str:
    return re.sub(r"\W", "_", index_name)


@click.command(
    "explore", help="Pipeline predictions over a data source for result exploration"
)
@click.argument("pipeline_path", type=Path(exists=True))
@click.option(
    "-ds", "--data-source", "data_source", type=Path(exists=True), required=True
)
@click.option("-e", "--explain", "explain", is_flag=True, default=False)
@click.option("-es", "--es-host", "es_host", envvar=ES_HOST, default=DEFAULT_ES_HOST)
def explore(pipeline_path: str, data_source: str, explain: bool, es_host: str) -> None:
    Pipeline.from_pretrained(pipeline_path).explore(
        data_source=DataSource.from_yaml(data_source), es_host=es_host, explain=explain
    )



from typing import Any, Dict

import numpy
import pytest
from allennlp.data import Instance

from biome.text import Pipeline, PipelineConfiguration
from biome.text.backbone import ModelBackbone
from biome.text.configuration import FeaturesConfiguration
from biome.text.modules.heads import TaskHeadSpec, TextClassification


class TestHead(TextClassification):
    def __init__(self, backbone: ModelBackbone):
        super(TestHead, self).__init__(backbone, labels=["test", "notest"])

    def explain_prediction(
        self, prediction: Dict[str, numpy.array], instance: Instance, n_steps: int
    ) -> Dict[str, Any]:
        return {}


class TestHeadWithRaise(TestHead):
    def explain_prediction(
        self, prediction: Dict[str, numpy.array], instance: Instance, n_steps: int
    ) -> Dict[str, Any]:
        raise NotImplementedError


def test_explain_tokenized_as_default():
    pipeline_config = PipelineConfiguration(
        name="test-classifier",
        features=FeaturesConfiguration(),
        head=TaskHeadSpec(type=TestHead),
    )
    pipeline = Pipeline.from_config(pipeline_config)
    prediction = pipeline.explain("This is a simple test with only tokens in explain")
    explain = prediction["explain"]

    assert explain
    assert explain.get("text")
    for token_info in explain["text"]:
        assert isinstance(token_info.get("token"), str)
        assert token_info.get("attribution") == 0.0


def test_explain_without_steps():
    pipeline_config = PipelineConfiguration(
        name="test-classifier",
        features=FeaturesConfiguration(),
        head=TaskHeadSpec(type=TestHeadWithRaise),
    )
    pipeline = Pipeline.from_config(pipeline_config)
    with pytest.raises(NotImplementedError):
        pipeline.explain("This is a simple test with only tokens in explain")

    prediction = pipeline.explain(
        "This is a simple test with only tokens in explain", n_steps=0
    )
    assert "explain" in prediction



from typing import Dict

import pandas as pd
import pytest
import yaml
from biome.text import TrainerConfiguration, VocabularyConfiguration
from biome.text import Pipeline
from biome.text.data import DataSource


@pytest.fixture
def training_data_source(tmp_path) -> DataSource:
    data_file = tmp_path / "record_pairs.json"
    df = pd.DataFrame(
        {
            "record1": [
                {"@fist_name": "Hans", "@last_name": "Peter"},
                {"@fist_name": "Heinrich", "@last_name": "Meier"},
                {"@fist_name": "Hans", "@last_name": "Peter"},
            ],
            "record2": [
                {"@fist_name": "Hans", "@last_name": "Petre"},
                {"@fist_name": "Heinz", "@last_name": "Meier"},
                {"@fist_name": "Hansel", "@last_name": "Peter"},
            ],
            "label": ["duplicate", "not_duplicate", "duplicate"],
        }
    )
    df.to_json(data_file, lines=True, orient="records")

    return DataSource(
        source=str(data_file), flatten=False, lines=True, orient="records"
    )


@pytest.fixture
def path_to_pipeline_yaml(tmp_path) -> str:
    pipeline_dict = {
        "name": "biome-bimpm",
        "tokenizer": {"text_cleaning": {"rules": ["strip_spaces"]}},
        "features": {
            # "tokens": {
            #     "indexer": {"type": "single_id", "lowercase_tokens": True},
            #     "embedder": {
            #         "type": "embedding",
            #         "embedding_dim": 32,
            #         "padding_index": 0,
            #     },
            # },
            "word": {
                "embedding_dim": 2,
                "lowercase_tokens": True,
                "trainable": True,
                "embedder": {"padding_index": 0},
            },
            # "token_characters": {
            #     "indexer": {
            #         "type": "characters",
            #         "character_tokenizer": {"lowercase_characters": True},
            #     },
            #     "embedder": {
            #         "type": "character_encoding",
            #         "embedding": {"embedding_dim": 32, "padding_index": 0},
            #         "encoder": {
            #             "type": "gru",
            #             "input_size": 32,
            #             "hidden_size": 112,
            #             "num_layers": 1,
            #             "bidirectional": True,
            #         },
            #         "dropout": 0.2,
            #     },
            # },
            "char": {
                "embedding_dim": 2,
                "dropout": 0.1,
                "encoder": {
                    "type": "gru",
                    "hidden_size": 2,
                    "num_layers": 1,
                    "bidirectional": True,
                },
                "indexer": {"character_tokenizer": {"lowercase_characters": True}},
                "embedder": {"embedding": {"padding_index": 0}},
            },
        },
        "head": {
            "type": "RecordPairClassification",
            "labels": ["duplicate", "not_duplicate"],
            "dropout": 0.1,
            "field_encoder": {
                "type": "gru",
                "bidirectional": False,
                # "input_size": 4,
                "hidden_size": 4,
                "num_layers": 1,
            },
            "record_encoder": {
                "type": "gru",
                "bidirectional": True,
                # "input_size": 4,
                "hidden_size": 2,
                "num_layers": 1,
            },
            "matcher_forward": {
                "is_forward": True,
                # "hidden_dim": 2,
                "num_perspectives": 10,
                "with_full_match": False,
            },
            "matcher_backward": {
                "is_forward": False,
                # "hidden_dim": 2,
                "num_perspectives": 10,
            },
            "aggregator": {
                "type": "gru",
                "bidirectional": True,
                # "input_size": ??,
                "hidden_size": 2,
                "num_layers": 1,
            },
            "classifier_feedforward": {
                # "input_dim": 8,
                "num_layers": 1,
                "hidden_dims": [4],
                "activations": ["relu"],
                "dropout": [0.1],
            },
            "initializer": {
                "regexes": [
                    ["_output_layer.weight", {"type": "xavier_normal"}],
                    ["_output_layer.bias", {"type": "constant", "val": 0}],
                    [".*linear_layers.*weight", {"type": "xavier_normal"}],
                    [".*linear_layers.*bias", {"type": "constant", "val": 0}],
                    [".*weight_ih.*", {"type": "xavier_normal"}],
                    [".*weight_hh.*", {"type": "orthogonal"}],
                    [".*bias.*", {"type": "constant", "val": 0}],
                    [".*matcher.*match_weights.*", {"type": "kaiming_normal"}],
                ]
            },
        },
    }
    pipeline_yaml = tmp_path / "pipeline.yml"
    with pipeline_yaml.open("w") as f:
        yaml.safe_dump(pipeline_dict, f)

    return str(pipeline_yaml)


@pytest.fixture
def trainer_dict() -> Dict:
    trainer_dict = {
        "num_epochs": 1,
        "optimizer": {"type": "adam", "amsgrad": True, "lr": 0.002},
    }

    return trainer_dict


def test_explain(path_to_pipeline_yaml):
    pipeline = Pipeline.from_yaml(path_to_pipeline_yaml)
    explain = pipeline.explain(
        record1={"first_name": "Hans"}, record2={"first_name": "Hansel"},
    )
    assert len(explain["explain"]["record1"]) == len(explain["explain"]["record2"])


def test_train(path_to_pipeline_yaml, training_data_source, trainer_dict, tmp_path):
    pipeline = Pipeline.from_yaml(path_to_pipeline_yaml,)
    pipeline.predict(record1={"first_name": "Hans"}, record2={"first_name": "Hansel"})
    pipeline.create_vocabulary(VocabularyConfiguration(sources=[training_data_source]))

    pipeline.train(
        output=str(tmp_path / "record_bimpm_experiment"),
        trainer=TrainerConfiguration(**trainer_dict),
        training=training_data_source,
        validation=training_data_source,
    )



from typing import Any, List, Optional, Union

from astroid import Instance

from biome.text import Pipeline, PipelineConfiguration
from biome.text.configuration import FeaturesConfiguration
from biome.text.modules.heads import TaskHeadSpec, TextClassification


class MyCustomHead(TextClassification):
    """Just a head renaming the original TextClassification head"""

    def inputs(self) -> Optional[List[str]]:
        return ["text", "second_text"]

    def featurize(
        self,
        text: Any,
        second_text: Optional[Any] = None,
        label: Optional[Union[int, str, List[Union[int, str]]]] = None,
    ) -> Optional[Instance]:
        instance = self.backbone.featurizer(
            {"text": text, "text-2": second_text},
            to_field=self.forward_arg_name,
            aggregate=True,
            exclude_record_keys=True,
        )
        return self.add_label(instance, label, to_field=self.label_name)


def test_check_pipeline_inputs_and_output():
    config = PipelineConfiguration(
        "test-pipeline",
        features=FeaturesConfiguration(),
        head=TaskHeadSpec(
            type=MyCustomHead,
            labels=[
                "blue-collar",
                "technician",
                "management",
                "services",
                "retired",
                "admin.",
            ],
        ),
    )

    pipeline = Pipeline.from_config(config)

    assert pipeline.inputs == ["text", "second_text"]
    assert pipeline.output == "label"



from typing import Any, Dict, List, Optional, Tuple

import numpy as np
import torch
from allennlp.data import Batch, Instance, TextFieldTensors
from allennlp.modules import (
    FeedForward,
    Seq2SeqEncoder,
    Seq2VecEncoder,
)
from allennlp.modules.bimpm_matching import BiMpmMatching
from allennlp.nn import InitializerApplicator, util
from captum.attr import IntegratedGradients

from biome.text.backbone import ModelBackbone
from biome.text.configuration import CharFeatures, WordFeatures
from biome.text.helpers import (
    get_char_tokens_ids_from_text_field_tensors,
    get_word_tokens_ids_from_text_field_tensors,
)
from biome.text.modules.encoders import TimeDistributedEncoder
from biome.text.modules.specs import (
    BiMpmMatchingSpec,
    FeedForwardSpec,
    Seq2SeqEncoderSpec,
    Seq2VecEncoderSpec,
)
from .classification import ClassificationHead
from ..task_head import TaskOutput


class RecordPairClassification(ClassificationHead):
    """Classifies the relation between a pair of records using a matching layer.

    The input for models using this `TaskHead` are two *records* with one or more *data fields* each, and a label
    describing their relationship.
    If you would like a meaningful explanation of the model's prediction,
    both records must consist of the same number of *data fields* and hold them in the same order.

    The architecture is loosely based on the AllenNLP implementation of the BiMPM model described in
    `Bilateral Multi-Perspective Matching for Natural Language Sentences <https://arxiv.org/abs/1702.03814>`_
    by Zhiguo Wang et al., 2017, and was adapted to deal with record pairs.

    Parameters
    ----------
    backbone : `ModelBackbone`
        Takes care of the embedding and optionally of the language encoding
    labels : `List[str]`
        List of labels
    field_encoder : `Seq2VecEncoder`
        Encodes a data field, contextualized within the field
    record_encoder : `Seq2SeqEncoder`
        Encodes data fields, contextualized within the record
    matcher_forward : `BiMPMMatching`
        BiMPM matching for the forward output of the record encoder layer
    matcher_backward : `BiMPMMatching`, optional
        BiMPM matching for the backward output of the record encoder layer
    aggregator : `Seq2VecEncoder`
        Aggregator of all BiMPM matching vectors
    classifier_feedforward : `FeedForward`
        Fully connected layers for classification.
        A linear output layer with the number of labels at the end will be added automatically!!!
    dropout : ``float``, optional (default=0.1)
        Dropout percentage to use.
    initializer : ``InitializerApplicator``, optional (default=``InitializerApplicator()``)
        If provided, will be used to initialize the model parameters.
    """

    def __init__(
        self,
        backbone: ModelBackbone,
        labels: List[str],
        field_encoder: Seq2VecEncoderSpec,
        record_encoder: Seq2SeqEncoderSpec,
        matcher_forward: BiMpmMatchingSpec,
        aggregator: Seq2VecEncoderSpec,
        classifier_feedforward: FeedForwardSpec,
        matcher_backward: BiMpmMatchingSpec = None,
        dropout: float = 0.1,
        initializer: InitializerApplicator = InitializerApplicator(),
    ):
        super(RecordPairClassification, self).__init__(backbone, labels)

        # This is needed for the TrainerConfig to choose the right 'sorting_keys'
        self.backbone.encoder = TimeDistributedEncoder(self.backbone.encoder)

        self._field_encoder: Seq2VecEncoder = field_encoder.input_dim(
            self.backbone.encoder.get_output_dim()
        ).compile()
        self._field_encoder = TimeDistributedEncoder(self._field_encoder)

        self._record_encoder: Seq2SeqEncoder = record_encoder.input_dim(
            self._field_encoder.get_output_dim()
        ).compile()

        input_dim_matcher = self._record_encoder.get_output_dim()
        if matcher_backward is not None:
            input_dim_matcher //= 2

        self._matcher_forward: BiMpmMatching = matcher_forward.input_dim(
            input_dim_matcher
        ).compile()
        self._matcher_backward = (
            matcher_backward.input_dim(input_dim_matcher).compile()
            if matcher_backward is not None
            else None
        )

        matching_dim = self._matcher_forward.get_output_dim()
        if self._matcher_backward:
            matching_dim += self._matcher_backward.get_output_dim()

        self._aggregator: Seq2VecEncoder = aggregator.input_dim(matching_dim).compile()

        # we aggregate the two records in the end -> *2
        self._classifier_feedforward: FeedForward = classifier_feedforward.input_dim(
            self._aggregator.get_output_dim() * 2
        ).compile()

        self._output_layer = torch.nn.Linear(
            self._classifier_feedforward.get_output_dim(), self.num_labels
        )

        self._dropout = torch.nn.Dropout(dropout)

        initializer(self)

    def featurize(
        self,
        record1: Dict[str, Any],
        record2: Dict[str, Any],
        label: Optional[str] = None,
    ) -> Optional[Instance]:
        """Tokenizes, indexes and embeds the two records and optionally adds the label

        Parameters
        ----------
        record1 : Dict[str, Any]
            First record
        record2 : Dict[str, Any]
            Second record
        label : Optional[str]
            Classification label

        Returns
        -------
        instance
            AllenNLP instance containing the two records plus optionally a label
        """
        record1_instance = self.backbone.featurizer(
            record1, to_field="record", aggregate=False
        )
        record2_instance = self.backbone.featurizer(
            record2, to_field="record", aggregate=False
        )
        instance = Instance(
            {
                "record1": record1_instance.get("record"),
                "record2": record2_instance.get("record"),
            }
        )
        instance = self.add_label(instance, label)

        return instance

    def forward(
        self,  # type: ignore
        record1: TextFieldTensors,
        record2: TextFieldTensors,
        label: torch.LongTensor = None,
    ) -> TaskOutput:
        # pylint: disable=arguments-differ
        """
        Parameters
        ----------
        record1
            Tokens of the first record.
            The dictionary is the output of a `ListField.as_array()`. It gives names to the tensors created by
            the `TokenIndexer`s.
            In its most basic form, using a `SingleIdTokenIndexer`, the dictionary is composed of:
            `{"tokens": {"tokens": Tensor(batch_size, num_fields, num_tokens)}}`.
            The dictionary is designed to be passed on directly to a `TextFieldEmbedder`, that has a
            `TokenEmbedder` for each key in the dictionary (except you set `allow_unmatched_keys` in the
            `TextFieldEmbedder` to False) and knows how to combine different word/character representations into a
            single vector per token in your input.
        record2
            Tokens of the second record.
        label : torch.LongTensor, optional (default = None)
            A torch tensor representing the sequence of integer gold class label of shape
            `(batch_size, num_classes)`.

        Returns
        -------
        An output dictionary consisting of:
        logits : torch.FloatTensor
        class_probabilities : torch.FloatTensor
        loss : torch.FloatTensor, optional
            A scalar loss to be optimised.
        """
        # embeding and encoding (field context)
        field_encoded_record1, record_mask_record1 = self._field_encoding(record1)
        field_encoded_record2, record_mask_record2 = self._field_encoding(record2)

        # encoding (record context), matching, aggregation, classification
        logits = self._bimpm_forward(
            field_encoded_record1,
            field_encoded_record2,
            record_mask_record1,
            record_mask_record2,
        )

        output: TaskOutput = self.single_label_output(logits, label)

        return output

    def _field_encoding(
        self, record: TextFieldTensors,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Embeds and encodes the records in a field context.

        We do this in a helper function to reuse it for the `explain_prediction` method.

        Parameters
        ----------
        record
            See `self.forward`

        Returns
        -------
        field_encoded_record
        record_mask
        """
        field_mask = util.get_text_field_mask(record, num_wrapping_dims=1)

        # embedding and encoding of one record (field context)
        embedded_record = self._dropout(
            self.backbone.forward(record, mask=field_mask, num_wrapping_dims=1)
        )
        field_encoded_record = self._dropout(
            self._field_encoder(embedded_record, mask=field_mask)
        )

        # mask for the record encoder
        record_mask = torch.sum(field_mask, -1) > 0

        return field_encoded_record, record_mask

    def _bimpm_forward(
        self,
        field_encoded_record1: torch.Tensor,
        field_encoded_record2: torch.Tensor,
        record_mask_record1: torch.Tensor,
        record_mask_record2: torch.Tensor,
    ) -> torch.Tensor:
        """Encodes in a record context, matches encodings, aggregates the matching results, classifies.

        We do this in a helper function to reuse it for the `explain_prediction` method.

        Parameters
        ----------
        field_encoded_record1
            Encoded record1 (in a field context)
        field_encoded_record2
            Encoded record2 (in a field context)
        record_mask_record1
            Mask for the record encoder for record1
        record_mask_record2
            Mask for the record encoder for record2

        Returns
        -------
        logits
        """
        record_encoded_record1 = self._dropout(
            self._record_encoder(field_encoded_record1, mask=record_mask_record1)
        )

        record_encoded_record2 = self._dropout(
            self._record_encoder(field_encoded_record2, mask=record_mask_record2)
        )

        # matching layer
        matching_vector_record1, matching_vector_record2 = self._matching_layer(
            record_encoded_record1,
            record_encoded_record2,
            record_mask_record1,
            record_mask_record2,
        )

        # aggregation layer
        aggregated_record1 = self._dropout(
            self._aggregator(matching_vector_record1, record_mask_record1)
        )
        aggregated_record2 = self._dropout(
            self._aggregator(matching_vector_record2, record_mask_record2)
        )

        aggregated_records = torch.cat([aggregated_record1, aggregated_record2], dim=-1)

        # the final feed forward layer
        logits = self._output_layer(self._classifier_feedforward(aggregated_records))

        return logits

    def _matching_layer(
        self,
        encoded_record1: torch.Tensor,
        encoded_record2: torch.Tensor,
        mask_record1: torch.Tensor,
        mask_record2: torch.Tensor,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Implements the Matching layer

        Parameters
        ----------
        encoded_record1
            encoded record1 (record context)
        encoded_record2
            encoded record2 (record context)
        mask_record1
            mask for record encoder for record1
        mask_record2
            mask for record encoder for record2

        Returns
        -------
        matching_vector_record1_cat
            Concatenated matching results for record1
        matching_vector_record2_cat
            Concatenated matching results for record2
        """
        matching_vector_record1: List[torch.Tensor] = []
        matching_vector_record2: List[torch.Tensor] = []

        # Matching is done separately for forward and backward encodings (if backward is specified)
        half_hidden_size = None
        if self._matcher_backward is not None:
            half_hidden_size = self._record_encoder.get_output_dim() // 2

        # forward matching
        matching_result = self._matcher_forward(
            encoded_record1[:, :, :half_hidden_size],
            mask_record1,
            encoded_record2[:, :, :half_hidden_size],
            mask_record2,
        )
        matching_vector_record1.extend(matching_result[0])
        matching_vector_record2.extend(matching_result[1])

        # backward matching
        if self._matcher_backward is not None:
            matching_result = self._matcher_backward(
                encoded_record1[:, :, half_hidden_size:],
                mask_record1,
                encoded_record2[:, :, half_hidden_size:],
                mask_record2,
            )
            matching_vector_record1.extend(matching_result[0])
            matching_vector_record2.extend(matching_result[1])

        # concat the matching vectors + apply dropout
        matching_vector_record1_cat = self._dropout(
            torch.cat(matching_vector_record1, dim=2)
        )
        matching_vector_record2_cat = self._dropout(
            torch.cat(matching_vector_record2, dim=2)
        )

        return matching_vector_record1_cat, matching_vector_record2_cat

    def explain_prediction(
        self, prediction: Dict[str, np.array], instance: Instance, n_steps: int
    ) -> Dict[str, Any]:
        """Calculates attributions for each data field in the record by integrating the gradients.

        IMPORTANT: The calculated attributions only make sense for a duplicate/not_duplicate binary classification task
        of the two records.

        TODO: optimize: for the prediction we already embedded and field encoded the records.
            Also, the forward passes here are always done on cpu!

        Parameters
        ----------
        prediction
        instance
        n_steps

        Returns
        -------
        prediction_dict
            The prediction dictionary with a newly added "explain" key
        """
        batch = Batch([instance])
        tokens_ids = batch.as_tensor_dict()

        # get attributions for each field
        field_encoded_record1, record_mask_record1 = self._field_encoding(
            tokens_ids.get("record1")
        )
        field_encoded_record2, record_mask_record2 = self._field_encoding(
            tokens_ids.get("record2")
        )
        if not field_encoded_record2.size() == field_encoded_record2.size():
            raise RuntimeError("Both records must have the same number of data fields!")

        ig = IntegratedGradients(self._bimpm_forward)

        prediction_target = int(np.argmax(prediction["probs"]))
        ig_attribute_record1 = ig.attribute(
            inputs=(field_encoded_record1, field_encoded_record2),
            baselines=(field_encoded_record2, field_encoded_record2),
            additional_forward_args=(record_mask_record1, record_mask_record2),
            target=prediction_target,
            return_convergence_delta=True,
            n_steps=n_steps,
        )

        ig_attribute_record2 = ig.attribute(
            inputs=(field_encoded_record1, field_encoded_record2),
            baselines=(field_encoded_record1, field_encoded_record1),
            additional_forward_args=(record_mask_record1, record_mask_record2),
            target=prediction_target,
            return_convergence_delta=True,
            n_steps=n_steps,
        )
        # The code below was an attempt to make attributions for the "duplicate case" more meaningful ... did not work
        # # duplicate case:
        # # Here we integrate each record along the path from the null vector -> record1/2
        # # assuming that the null vector provides the highest "not duplicate" score.
        # if prediction_target == 0:
        #     ig_attribute_record1 = ig.attribute(
        #         inputs=(field_encoded_record1, field_encoded_record2),
        #         baselines=(torch.zeros_like(field_encoded_record1), field_encoded_record2),
        #         additional_forward_args=(record_mask_record1, record_mask_record2),
        #         # we fix the target since we want negative integrals always to be associated
        #         # to the "not_duplicate" case and positive ones to the "duplicate" case
        #         target=0,
        #         return_convergence_delta=True,
        #     )
        #
        #     ig_attribute_record2 = ig.attribute(
        #         inputs=(field_encoded_record1, field_encoded_record2),
        #         baselines=(field_encoded_record1, torch.zeros_like(field_encoded_record2)),
        #         additional_forward_args=(record_mask_record1, record_mask_record2),
        #         # we fix the target since we want negative integrals always to be associated
        #         # to the "not_duplicate" case and positive ones to the "duplicate" case
        #         target=0,
        #         return_convergence_delta=True,
        #     )
        #
        # # not duplicate case:
        # # Here we integrate each record along the path from record2/1 -> record1/2
        # # assuming that the same record provides the highest "duplicate" score.
        # elif prediction_target == 1:
        #     ...
        # else:
        #     raise RuntimeError("The `explain` method is only implemented for a binary classification task: "
        #                        "[duplicate, not_duplicate]")

        attributions_record1, delta_record1 = self._get_attributions_and_delta(
            ig_attribute_record1, 0
        )
        attributions_record2, delta_record2 = self._get_attributions_and_delta(
            ig_attribute_record2, 1
        )

        # get tokens corresponding to the attributions
        field_tokens_record1 = self._get_field_tokens(tokens_ids.get("record1"))
        field_tokens_record2 = self._get_field_tokens(tokens_ids.get("record2"))

        return {
            **prediction,
            "explain": {
                "record1": [
                    {"token": token, "attribution": attribution}
                    for token, attribution in zip(
                        field_tokens_record1, attributions_record1
                    )
                ],
                "record2": [
                    {"token": token, "attribution": attribution}
                    for token, attribution in zip(
                        field_tokens_record2, attributions_record2
                    )
                ],
            },
        }

    def _get_field_tokens(self, record_token_ids: TextFieldTensors) -> List[str]:
        """
        TODO: This can very likely be optimised!
        Parameters
        ----------
        record_token_ids

        Returns
        -------

        """
        field_tokens = []

        if WordFeatures.namespace in record_token_ids:
            # batch size is 1 -> [0]
            for field in get_word_tokens_ids_from_text_field_tensors(record_token_ids)[
                0
            ]:
                tokens = []
                for word_idx in field:
                    # skipp padding
                    if word_idx.item() == 0:
                        continue
                    token = self.backbone.vocab.get_token_from_index(
                        word_idx.item(), namespace=WordFeatures.namespace
                    )
                    tokens.append(token)
                field_tokens.append(" ".join(tokens))

        elif CharFeatures.namespace in record_token_ids:
            # batch size is 1 -> [0]
            for field in get_char_tokens_ids_from_text_field_tensors(record_token_ids)[
                0
            ]:
                tokens = []
                for word in field:
                    token = []
                    for char_idx in word:
                        # skipp padding
                        if char_idx.item() == 0:
                            continue
                        char = self.backbone.vocab.get_token_from_index(
                            char_idx.item(), namespace=CharFeatures.namespace
                        )
                        token.append(char)
                    if token:
                        tokens.append("".join(token))
                field_tokens.append(" ".join(tokens))

        return field_tokens

    @staticmethod
    def _get_attributions_and_delta(
        ig_attribute_output, zero_or_one: int
    ) -> Tuple[np.array, float]:
        """Gets attributions and delta out of the `IntegratedGradients.attribute()` output.

        Parameters
        ----------
        ig_attribute_output
            Output of the `IntegratedGradients.attribute()` method.
        zero_or_one
            Is either 0 for record1 or 1 for record2

        Returns
        -------
        attributions, delta
        """
        attributions = ig_attribute_output[0][zero_or_one].sum(dim=2).squeeze(0)
        attributions = attributions / torch.norm(attributions)
        attributions = attributions.cpu().detach().numpy()
        delta = ig_attribute_output[1]

        return attributions, delta



import unittest

from distributed import LocalCluster

from biome.text.data.helpers import (
    configure_dask_cluster,
    close_dask_client,
)


class UtilTests(unittest.TestCase):
    def test_dask_cluster_local(self):

        client = configure_dask_cluster()
        cluster = client.cluster

        self.assertTrue(
            isinstance(cluster, LocalCluster), msg=f"Wrong cluster type: {cluster}"
        )
        self.assertEqual("running", client.status)
        self.assertEqual("running", cluster.status)

        client = configure_dask_cluster(address="localhost:8786")
        self.assertEqual(client, configure_dask_cluster())

        close_dask_client()
        self.assertEqual("closed", client.status)
        self.assertEqual("closed", cluster.status)

    def test_dask_cluster_client(self):
        port = 8788
        cluster = LocalCluster(processes=False, scheduler_port=port)

        client = configure_dask_cluster(address=f"localhost:{port}")
        self.assertEqual(None, client.cluster)
        self.assertEqual("running", client.status)

        close_dask_client()
        self.assertEqual("closed", client.status)
        self.assertEqual("running", cluster.status)

        cluster.close(timeout=10)
        self.assertEqual("closed", cluster.status)



from typing import Any, Dict, Optional

from biome.text.modules.specs import Seq2VecEncoderSpec


class WordFeatures:
    """Feature configuration at word level"""

    namespace = "word"

    def __init__(
        self,
        embedding_dim: int,
        lowercase_tokens: bool = False,
        trainable: bool = True,
        weights_file: Optional[str] = None,
        **extra_params
    ):
        self.embedding_dim = embedding_dim
        self.lowercase_tokens = lowercase_tokens
        self.trainable = trainable
        self.weights_file = weights_file
        self.extra_params = extra_params

    @property
    def config(self):
        config = {
            "indexer": {
                "type": "single_id",
                "lowercase_tokens": self.lowercase_tokens,
                "namespace": self.namespace,
            },
            "embedder": {
                "embedding_dim": self.embedding_dim,
                "vocab_namespace": self.namespace,
                "trainable": self.trainable,
                **({"pretrained_file": self.weights_file} if self.weights_file else {}),
            },
        }

        for k in self.extra_params:
            config[k] = {**self.extra_params[k], **config.get(k)}

        return config

    def to_json(self):
        data = vars(self)
        data.update(data.pop("extra_params"))

        return data

    def to_dict(self):
        return {
            "embedding_dim": self.embedding_dim,
            "lowercase_tokens": self.lowercase_tokens,
            "trainable": self.trainable,
            "weights_file": self.weights_file,
            **self.extra_params,
        }


class CharFeatures:
    """Feature configuration at character level"""

    namespace = "char"

    def __init__(
        self,
        embedding_dim: int,
        encoder: Dict[str, Any],
        dropout: float = 0.0,
        lowercase_characters: bool = False,
        **extra_params
    ):
        self.embedding_dim = embedding_dim
        self.encoder = encoder
        self.dropout = dropout
        self.lowercase_characters = lowercase_characters
        self.extra_params = extra_params

    @property
    def config(self):
        config = {
            "indexer": {
                "type": "characters",
                "namespace": self.namespace,
                "character_tokenizer": {
                    "lowercase_characters": self.lowercase_characters
                },
            },
            #         "character_tokenizer": {"lowercase_characters": True},
            "embedder": {
                "type": "character_encoding",
                "embedding": {
                    "embedding_dim": self.embedding_dim,
                    "vocab_namespace": self.namespace,
                },
                "encoder": Seq2VecEncoderSpec(**self.encoder)
                .input_dim(self.embedding_dim)
                .config,
                "dropout": self.dropout,
            },
        }

        for k, v in self.extra_params.items():
            config[k] = {**self.extra_params[k], **config.get(k)}

        return config

    def to_json(self):
        data = vars(self)
        data.update(data.pop("extra_params"))

        return data

    def to_dict(self):
        return {
            "embedding_dim": self.embedding_dim,
            "encoder": self.encoder,
            "dropout": self.dropout,
            **self.extra_params,
        }



import pytest
import yaml

from biome.text import Pipeline
from biome.text.configuration import (
    CharFeatures,
    FeaturesConfiguration,
    PipelineConfiguration,
    TaskHeadSpec,
    TokenizerConfiguration,
    WordFeatures,
)
from biome.text.modules.heads import TextClassification
from biome.text.modules.specs.allennlp_specs import Seq2SeqEncoderSpec


@pytest.fixture
def pipeline_yaml(tmp_path):
    pipeline_dict = {
        "name": "test_pipeline_config",
        "tokenizer": {"text_cleaning": {"rules": ["strip_spaces"]},},
        "features": {
            "word": {"embedding_dim": 2, "lowercase_tokens": True,},
            "char": {
                "embedding_dim": 2,
                "encoder": {
                    "type": "gru",
                    "hidden_size": 2,
                    "num_layers": 1,
                    "bidirectional": True,
                },
                "dropout": 0.1,
            },
        },
        "encoder": {
            "type": "gru",
            "hidden_size": 2,
            "num_layers": 1,
            "bidirectional": True,
        },
        "head": {
            "type": "TextClassification",
            "labels": ["duplicate", "not_duplicate"],
            "pooler": {"type": "boe",},
        },
    }

    yaml_path = tmp_path / "pipeline.yml"
    with yaml_path.open("w") as file:
        yaml.safe_dump(pipeline_dict, file)

    return str(yaml_path)


def test_pipeline_without_word_features():
    tokenizer_config = TokenizerConfiguration()
    char_features = CharFeatures(
        embedding_dim=2,
        encoder={
            "type": "gru",
            "hidden_size": 2,
            "num_layers": 1,
            "bidirectional": True,
        },
        dropout=0.1,
    )
    features_config = FeaturesConfiguration(char=char_features)
    encoder_spec = Seq2SeqEncoderSpec(
        type="gru", hidden_size=2, num_layers=1, bidirectional=True
    )

    head_spec = TaskHeadSpec(
        type="TextClassification",
        labels=["duplicate", "not_duplicate"],
        pooler={"type": "boe"},
    )

    pipeline_config = PipelineConfiguration(
        name="no_word_features",
        tokenizer=tokenizer_config,
        features=features_config,
        encoder=encoder_spec,
        head=head_spec,
    )

    pl = Pipeline.from_config(pipeline_config)
    assert "word" not in pl.backbone.featurizer.indexer
    assert "char" in pl.backbone.featurizer.indexer


def test_pipeline_config(pipeline_yaml):
    tokenizer_config = TokenizerConfiguration(text_cleaning={"rules": ["strip_spaces"]})

    word_features = WordFeatures(embedding_dim=2, lowercase_tokens=True)
    char_features = CharFeatures(
        embedding_dim=2,
        encoder={
            "type": "gru",
            "hidden_size": 2,
            "num_layers": 1,
            "bidirectional": True,
        },
        dropout=0.1,
    )
    features_config = FeaturesConfiguration(word=word_features, char=char_features)

    encoder_spec = Seq2SeqEncoderSpec(
        type="gru", hidden_size=2, num_layers=1, bidirectional=True
    )

    head_spec = TaskHeadSpec(
        type=TextClassification,
        labels=["duplicate", "not_duplicate"],
        pooler={"type": "boe"},
    )

    pipeline_config = PipelineConfiguration(
        name="test_pipeline_config",
        tokenizer=tokenizer_config,
        features=features_config,
        encoder=encoder_spec,
        head=head_spec,
    )

    pl = Pipeline.from_config(pipeline_config)

    pl_yaml = Pipeline.from_yaml(pipeline_yaml)

    assert pl.trainable_parameter_names == pl_yaml.trainable_parameter_names
    assert pl.trainable_parameters == pl_yaml.trainable_parameters

    sample_text = "My simple text"
    assert pl.backbone.featurizer(sample_text) == pl_yaml.backbone.featurizer(
        sample_text
    )



from typing import Dict, Optional

import torch
from allennlp.common.checks import ConfigurationError
from allennlp.data import Instance, TextFieldTensors
from allennlp.modules import SoftmaxLoss
from allennlp.nn.util import get_text_field_mask, get_token_ids_from_text_field_tensors
from allennlp.training.metrics import Perplexity

from biome.text import vocabulary
from biome.text.backbone import ModelBackbone
from biome.text.modules.specs import ComponentSpec
from .task_head import TaskHead, TaskName, TaskOutput


class LanguageModelling(TaskHead):
    """
    Task head for next-token language modelling, i.e., a model to predict the next token
    in a sequence of tokens.
    """

    def task_name(self) -> TaskName:
        return TaskName.language_modelling

    def __init__(self, backbone: ModelBackbone, dropout: float = None) -> None:
        super(LanguageModelling, self).__init__(backbone)

        if not backbone.featurizer.has_word_features:
            raise ConfigurationError(
                "`LanguageModelling` defines a word-level next token language model. "
                "Please check your `features` configuration to enable at least `words` features."
            )

        self._forward_dim = backbone.encoder.get_output_dim()

        if dropout:
            self._dropout = torch.nn.Dropout(dropout)
        else:
            self._dropout = lambda x: x

        self.metrics = {"perplexity": Perplexity()}

        self._loss = SoftmaxLoss(
            num_words=vocabulary.words_vocab_size(self.backbone.vocab),
            embedding_dim=self.backbone.encoder.get_output_dim(),
        )

    def on_vocab_update(self):
        self._loss = SoftmaxLoss(
            num_words=vocabulary.words_vocab_size(self.backbone.vocab),
            embedding_dim=self.backbone.encoder.get_output_dim(),
        )

    def featurize(self, text: str) -> Optional[Instance]:
        return self.backbone.featurizer(text, to_field="text", aggregate=True)

    def forward(  # type: ignore
        self, text: TextFieldTensors
    ) -> TaskOutput:

        mask = get_text_field_mask(text)
        contextual_embeddings = self.backbone.forward(text, mask)
        # NOTE: @dvsrepo, Allennlp 1.0 includes a second features level that I'm not sure of understand.
        # Anyway, they proved a function to realize the target here (the function docstring clarifies the
        # real spaghetti inside indexer code references, :-)
        token_ids = get_token_ids_from_text_field_tensors(text)
        assert isinstance(contextual_embeddings, torch.Tensor)

        # Use token_ids to compute targets
        # targets are next token ids with respect to first token in the seq
        # e.g. token_ids [[1, 3, 5, 7],..[]], forward_targets=[[3,5,7],..]
        forward_targets = torch.zeros_like(token_ids)
        forward_targets[:, 0:-1] = token_ids[:, 1:]

        # add dropout
        contextual_embeddings_with_dropout = self._dropout(contextual_embeddings)

        # compute softmax loss
        try:
            forward_loss = self._compute_loss(
                contextual_embeddings_with_dropout, forward_targets
            )
        except IndexError:
            raise IndexError(
                "Word token out of vocabulary boundaries, please check your vocab is correctly set"
                " or created before starting training."
            )

        num_targets = torch.sum((forward_targets > 0).long())
        if num_targets > 0:
            average_loss = forward_loss / num_targets.float()
        else:
            average_loss = torch.tensor(0.0).to(forward_targets.device)

        for metric in self.metrics.values():
            metric(average_loss)

        return TaskOutput(
            logits=None,
            probs=None,
            loss=average_loss,
            **{"lm_embeddings": contextual_embeddings, "mask": mask}
        )

    def get_metrics(self, reset: bool = False) -> Dict[str, float]:
        return {
            metric_name: metric.get_metric(reset)
            for metric_name, metric in self.metrics.items()
        }

    def _compute_loss(
        self, lm_embeddings: torch.Tensor, forward_targets: torch.Tensor
    ) -> torch.Tensor:
        forward_embeddings = lm_embeddings
        forward_loss = self._loss_helper(0, forward_embeddings, forward_targets)
        return forward_loss

    def _loss_helper(
        self,
        direction: int,
        direction_embeddings: torch.Tensor,
        direction_targets: torch.Tensor,
    ) -> torch.Tensor:
        mask = direction_targets > 0
        # we need to subtract 1 to undo the padding id since the softmax
        # does not include a padding dimension

        # shape (batch_size * timesteps, )
        non_masked_targets = direction_targets.masked_select(mask) - 1

        # shape (batch_size * timesteps, embedding_dim)
        non_masked_embeddings = direction_embeddings.masked_select(
            mask.unsqueeze(-1)
        ).view(-1, self._forward_dim)
        return self._loss(non_masked_embeddings, non_masked_targets)


class LanguageModellingSpec(ComponentSpec[LanguageModelling]):
    """Spec for language model head components"""

    pass



import pytest

from biome.text.commons import ImmutableDict


class TestImmutableDict:
    def test_cannot_mutate(self):
        dict = ImmutableDict(a=1, b="2", c=1000.00)

        with pytest.raises(TypeError):
            dict.f = "F"

        with pytest.raises(TypeError):
            dict.a = 100

    # TODO: Test a serialization/deserialization



import os

import pytest

from biome.text.data import DataSource
from biome.text.errors import MissingArgumentError

from tests import DaskSupportTest, RESOURCES_PATH

FILES_PATH = os.path.join(RESOURCES_PATH, "data")


class DataSourceTest(DaskSupportTest):
    def test_wrong_format(self):
        with pytest.raises(MissingArgumentError):
            DataSource(format="not-found")
        # New format
        with pytest.raises(TypeError):
            DataSource(source="not-found")

    def test_add_mock_format(self):
        def ds_parser(*args, **kwargs):
            from dask import dataframe as ddf
            import pandas as pd

            return ddf.from_pandas(
                pd.DataFrame([i for i in range(0, 100)]), npartitions=1
            )

        DataSource.add_supported_format("new-format", ds_parser)
        self.assertFalse(
            DataSource(source="source", format="new-format").to_dataframe().columns
            is None
        )

    def test_to_mapped(self):
        the_mapping = {"label": "overall", "tokens": "summary"}

        for ds in [
            DataSource(
                format="json",
                mapping=the_mapping,
                source=os.path.join(FILES_PATH, "dataset_source.jsonl"),
            ),
            DataSource(
                source=os.path.join(FILES_PATH, "dataset_source.jsonl"),
                mapping=the_mapping,
            ),
        ]:
            df = ds.to_mapped_dataframe()

            self.assertIn("label", df.columns)
            self.assertIn("tokens", df.columns)

    def test_no_mapping(self):

        ds = DataSource(
            format="json", source=os.path.join(FILES_PATH, "dataset_source.jsonl")
        )
        with pytest.raises(ValueError):
            ds.to_mapped_dataframe()

    def test_load_multiple_formats(self):
        files = [
            os.path.join(FILES_PATH, "dataset_source.jsonl"),
            os.path.join(FILES_PATH, "dataset_source.csv"),
        ]
        with pytest.raises(TypeError):
            DataSource(source=files)

    def test_override_format(self):
        with pytest.raises(TypeError):
            DataSource(source=os.path.join(FILES_PATH, "*.jsonl"), format="not-found")



import warnings
from typing import Any, Dict, List, Optional, Union

import torch
from allennlp.data import Instance, TextFieldTensors, Vocabulary
from allennlp.modules import TextFieldEmbedder
from allennlp.modules.seq2seq_encoders import PassThroughEncoder

from .featurizer import InputFeaturizer
from .modules.encoders import Encoder


class ModelBackbone(torch.nn.Module):
    """The backbone of the model.

    It is composed of a tokenizer, featurizer and an encoder.
    This component of the model can be pretrained and used with different task heads.

    Attributes
    ----------
    vocab: `Vocabulary`
        The vocabulary of the pipeline
    featurizer: `InputFeaturizer`
        Defines the input features of the tokens and indexes
    embedder: `TextFieldEmbedder`
        The embedding layer
    encoder: Encoder
        Outputs an encoded sequence of the tokens
    """

    def __init__(
        self,
        vocab: Vocabulary,
        featurizer: InputFeaturizer,
        embedder: TextFieldEmbedder,
        encoder: Optional[Encoder] = None,
    ):
        super(ModelBackbone, self).__init__()

        self.vocab = vocab
        self.featurizer = featurizer
        self.embedder = embedder
        self.encoder = (
            encoder.input_dim(self.embedder.get_output_dim()).compile()
            if encoder
            else PassThroughEncoder(self.embedder.get_output_dim())
        )

    def forward(
        self, text: TextFieldTensors, mask: torch.Tensor, num_wrapping_dims: int = 0,
    ) -> torch.Tensor:
        """Applies the embedding and encoding layer

        Parameters
        ----------
        text
            Output of the `batch.as_tensor_dict()` method, basically the indices of the indexed tokens
        mask
            A mask indicating which one of the tokens are padding tokens
        num_wrapping_dims
            0 if `text` is the output of a `TextField`, 1 if it is the output of a `ListField`

        Returns
        -------
        tensor
            Encoded representation of the input
        """
        embeddings = self.embedder(text, num_wrapping_dims=num_wrapping_dims)
        return self.encoder(embeddings, mask=mask)

    def on_vocab_update(self):
        """This method is called when a base model updates the vocabulary"""
        # This loop applies only for embedding layer.
        for model_path, module in self.named_modules():
            if hasattr(module, "extend_vocab"):
                module.extend_vocab(self.vocab)

    def featurize(
        self,
        record: Union[str, List[str], Dict[str, Any]],
        to_field: str = "record",
        aggregate: bool = False,
        tokenize: bool = True,
    ) -> Instance:
        """Generates an allennlp instance from a record input.
        DEPRECATED: use self.featurizer instead

        Parameters
        ----------
        record: `Union[str, List[str], Dict[str, Any]]`
            Input data
        to_field: `str`
            The field name in the returned instance
        aggregate: `bool`
            If true, the returned instance will contain a single `TextField` with all record fields;
            If false, the instance will contain a `ListField` of `TextField`s.
        tokenize: `bool`
            If false, skip tokenization phase and pass record data as tokenized token list.

        Returns
        -------
        instance: `Instance`
        """
        warnings.warn(
            "backbone.featurize is deprecated. Use instead backbone.featurizer",
            DeprecationWarning,
        )
        return self.featurizer(record, to_field, aggregate, tokenize)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import glob
import os
import sys
from typing import Tuple

try:
    from setuptools import setup, find_namespace_packages
except ImportError as error:
    raise ImportError("Make sure you have setuptools >= 40.1.0 installed!") from error

from pip import __version__ as pip_version

PIP_VERSION_REQUIRED = "19.1.1"


def check_pip_version(required_version: str, version: str) -> bool:
    def version_str_2_numbers(version: str) -> Tuple[int, int, int]:
        version_fractions = [int(n) for n in version.split(".")]
        return tuple(
            [
                version_fractions[i] if i < len(version_fractions) else 0
                for i in range(0, 3)
            ]
        )

    mayor, minor, fixes = version_str_2_numbers(version)
    req_mayor, req_minor, req_fixes = version_str_2_numbers(required_version)

    if (
        mayor > req_mayor
        or (mayor == req_mayor and minor > req_minor)
        or (mayor == req_mayor and minor == req_minor and fixes >= fixes)
    ):
        return True

    print(f"Minimal pip version should be {required_version}, found: {version}")
    print(f"Please upgrade pip: pip install --upgrade pip")
    return False


def about_info(package: str):
    """Fetch about info """
    root = os.path.abspath(os.path.dirname(__file__))
    with open(
        os.path.join(root, "src", package.replace("-", "/"), "about.py"),
        encoding="utf8",
    ) as f:
        about = {}
        exec(f.read(), about)
        return about


if __name__ == "__main__":
    if not check_pip_version(PIP_VERSION_REQUIRED, pip_version):
        sys.exit(1)

    package_name = "biome-text"
    about = about_info(package_name)

    setup(
        name=package_name,
        version=about["__version__"],
        description="Biome-text is a light-weight open source Natural Language Processing toolbox"
        " built with AllenNLP",
        author="Recognai",
        author_email="francisco@recogn.ai",
        url="https://www.recogn.ai/",
        long_description=open("README.md").read(),
        long_description_content_type="text/markdown",
        packages=find_namespace_packages("src"),
        package_dir={"": "src"},
        install_requires=[
            "allennlp~=1.0.0rc",
            "gevent~=1.4.0",
            "flask~=1.1.2",
            "flask-cors~=3.0.8",
            "click~=7.1.0",
            "beautifulsoup4~=4.9.0",
            "lxml~=4.5.0",
            "fastapi~=0.55.0",
            "uvicorn~=0.11.0",
            "dask[complete]~=2.17.0",
            "msgpack~=0.6.0",
            "cachey~=0.2.0",
            "pyarrow~=0.17.0",
            "ujson~=2.0.0",
            "pandas~=1.0.0",
            "dask-elk~=0.4.0",
            "bokeh~=2.0.0",
            "xlrd~=1.2.0",
            "flatdict~=4.0.0",
            "s3fs~=0.4.0",
            "captum~=0.2.0",
            "ipywidgets~=7.5.1",
        ],
        extras_require={
            "testing": [
                "pytest~=5.4.3",
                "pytest-cov~=2.10.0",
                "pytest-pylint~=0.14.0",
                "black",
                "GitPython",
                "pdoc3~=0.8.1",
                "pytest-notebook~=0.6.0",
            ]
        },
        package_data={
            "biome": [
                file.replace("src/biome/", "")
                for file in glob.glob("src/biome/text/ui/webapp/**/*.*", recursive=True)
            ]
        },
        entry_points={"console_scripts": ["biome=biome.text.cli:main",]},
        python_requires=">=3.6.1",  # taken from AllenNLP
        zip_safe=False,
    )



import json
import os
from logging import Logger

import requests
from flask import Flask, Response, jsonify, request, send_file, send_from_directory
from flask_cors import CORS
from werkzeug.exceptions import NotFound
from werkzeug.middleware.proxy_fix import ProxyFix


def make_app(es_host: str, statics_dir: str) -> Flask:
    app = Flask(__name__)  # pylint: disable=invalid-name
    app.wsgi_app = ProxyFix(
        app.wsgi_app
    )  # sets the requester IP with the X-Forwarded-For header
    CORS(app)

    logger: Logger = app.logger

    @app.route(
        "/elastic/"
    )  # Fixes the `WARNING:biome.text.ui.app:Path elastic/ not found` message"
    @app.route(
        "/elastic/<path:path>", methods=["GET", "OPTIONS", "POST", "PUT", "DELETE"]
    )
    def elasticsearch_proxy_handler(path: str = "") -> Response:
        """Handles elasticsearch requests"""
        es_url = f"{es_host}/{path}"

        response = None
        if request.method == "OPTIONS":
            return Response(response="", status=200)
        if request.method == "GET":
            response = requests.get(es_url)
        if request.method == "POST":
            response = requests.post(es_url, json=request.get_json())
        if request.method == "PUT":
            response = requests.put(es_url, json=request.get_json())
        if request.method == "DELETE":
            response = requests.delete(es_url, json=request.get_json())
        return (
            jsonify(json.loads(response.text))
            if response
            else Response(response="Not found", status=404)
        )

    @app.route("/<path:path>")
    def static_proxy(path: str) -> Response:  # pylint: disable=unused-variable
        try:
            return send_from_directory(statics_dir, path)
        except NotFound as error:
            logger.warning("Path %s not found. Error: %s", path, error)
            return index()

    @app.route("/static/js/<path:path>")
    def static_js_proxy(path: str) -> Response:  # pylint: disable=unused-variable
        return send_from_directory(os.path.join(statics_dir, "static/js"), path)

    @app.route("/static/css/<path:path>")
    def static_css_proxy(path: str) -> Response:  # pylint: disable=unused-variable
        return send_from_directory(os.path.join(statics_dir, "static/css"), path)

    @app.route("/public/<path:path>")
    def public_proxy(path: str) -> Response:  # pylint: disable=unused-variable
        return send_from_directory(statics_dir, path)

    @app.route("/")
    def index() -> Response:  # pylint: disable=unused-variable
        return send_file(os.path.join(statics_dir, "index.html"))

    return app



import os
from tempfile import mkdtemp

from biome.text import Pipeline, PipelineConfiguration
from biome.text.configuration import (
    FeaturesConfiguration,
    TrainerConfiguration,
    VocabularyConfiguration,
)
from biome.text.data import DataSource
from biome.text.modules.heads import TaskHeadSpec, TextClassification
from tests.test_context import TEST_RESOURCES


class MyCustomHead(TextClassification):
    """Just a head renaming the original TextClassification head"""

    pass


def test_load_pipeline_with_custom_head():
    config = PipelineConfiguration(
        "test-pipeline",
        features=FeaturesConfiguration(),
        head=TaskHeadSpec(
            type=MyCustomHead,
            labels=[
                "blue-collar",
                "technician",
                "management",
                "services",
                "retired",
                "admin.",
            ],
        ),
    )

    pipeline = Pipeline.from_config(config)
    assert isinstance(pipeline.head, MyCustomHead)

    train = DataSource(
        source=os.path.join(TEST_RESOURCES, "resources/data/dataset_source.csv"),
        mapping={"label": "job", "text": ["education", "marital"]},
    )
    output = mkdtemp()
    pipeline.create_vocabulary(VocabularyConfiguration(sources=[train]))
    pipeline.train(output=output, training=train)

    trained_pl = Pipeline.from_pretrained(os.path.join(output, "model.tar.gz"))
    trained_pl.predict("Oh yeah")
    assert isinstance(trained_pl.head, MyCustomHead)
